import warnings
from abc import ABC
from typing import Any, Tuple, Optional, List
from copy import deepcopy

import math
import numpy as np
from numpy.typing import ArrayLike
from scipy import interpolate
import matplotlib.pyplot as plt

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError):
    import numpy as cp

from . import detector as lisa_models
from .utils.utility import AET
from .utils.constants import *
from .stochastic import (
    StochasticContribution,
    FittedHyperbolicTangentGalacticForeground,
)

"""
The sensitivity code is heavily based on an original code by Stas Babak, Antoine Petiteau for the LDC team.

"""


class Sensitivity(ABC):
    """Base Class for PSD information."""

    @staticmethod
    def transform(
        f: float | np.ndarray,
        Spm: float | np.ndarray,
        Sop: float | np.ndarray,
        **kwargs: dict,
    ) -> float | np.ndarray:
        """Transform from the base sensitivity functions to the TDI PSDs.

        Args:
            f: Frequency array.
            Spm: Acceleration term.
            Sop: OMS term.
            **kwargs: For interoperability.

        Returns:
            Transformed TDI PSD values.

        """
        raise NotImplementedError

    @staticmethod
    def lisanoises(
        f: float | np.ndarray,
        model: Optional[lisa_models.LISAModel | str] = lisa_models.scirdv1,
        unit: Optional[str] = "relative_frequency",
    ) -> Tuple[float, float]:
        """Calculate both LISA noise terms based on input model.

        Args:
            f: Frequency array.
            model: Noise model. Object of type :class:`lisa_models.LISAModel`. It can also be a string corresponding to one of the stock models.
            unit: Either ``"relative_frequency"`` or ``"displacement"``.

        Returns:
            Tuple with acceleration term as first value and oms term as second value.

        """

        if isinstance(model, str):
            model = lisa_models.check_lisa_model(model)

        # TODO: fix this up
        Soms_d_in = model.Soms_d
        Sa_a_in = model.Sa_a

        frq = f
        ### Acceleration noise
        ## In acceleration
        Sa_a = Sa_a_in * (1.0 + (0.4e-3 / frq) ** 2) * (1.0 + (frq / 8e-3) ** 4)
        ## In displacement
        Sa_d = Sa_a * (2.0 * np.pi * frq) ** (-4.0)
        ## In relative frequency unit
        Sa_nu = Sa_d * (2.0 * np.pi * frq / C_SI) ** 2
        Spm = Sa_nu

        ### Optical Metrology System
        ## In displacement
        Soms_d = Soms_d_in * (1.0 + (2.0e-3 / f) ** 4)
        ## In relative frequency unit
        Soms_nu = Soms_d * (2.0 * np.pi * frq / C_SI) ** 2
        Sop = Soms_nu

        if unit == "displacement":
            return Sa_d, Soms_d
        elif unit == "relative_frequency":
            return Spm, Sop

    @classmethod
    def get_Sn(
        cls,
        f: float | np.ndarray,
        model: Optional[lisa_models.LISAModel | str] = lisa_models.scirdv1,
        **kwargs: dict,
    ) -> float | np.ndarray:
        """Calculate the PSD

        Args:
            f: Frequency array.
            model: Noise model. Object of type :class:`lisa_models.LISAModel`. It can also be a string corresponding to one of the stock models.
            **kwargs: For interoperability.

        Returns:
            PSD values.

        """
        x = 2.0 * np.pi * lisaLT * f

        # get noise values
        Spm, Sop = cls.lisanoises(f, model)

        # transform as desired for TDI combination
        Sout = cls.transform(f, Spm, Sop, **kwargs)

        # will add zero if ignored
        Sout += cls.get_stochastic_contribution(f, **kwargs)
        return Sout

    @classmethod
    def get_stochastic_contribution(
        cls,
        f: float | np.ndarray,
        stochastic_params: Optional[tuple] = (),
        stochastic_kwargs: Optional[dict] = {},
        stochastic_function: Optional[StochasticContribution | str] = None,
    ) -> float | np.ndarray:
        """Calculate contribution from stochastic signal.

        This function directs and wraps the calculation of and returns
        the stochastic signal. The ``stochastic_function`` calculates the
        sensitivity contribution. The ``transform_factor`` can transform that
        output to the correct TDI contribution.

        Args:
            f: Frequency array.
            stochastic_params: Parameters (arguments) to feed to ``stochastic_function``.
            stochastic_kwargs: Keyword arguments to feeed to ``stochastic_function``.
            stochastic_function: Stochastic class or string name of stochastic class. Takes ``stochastic_args`` and ``stochastic_kwargs``.
                If ``None``, it uses :class:`FittedHyperbolicTangentGalacticForeground`.

        Returns:
            Contribution from stochastic signal.


        """

        if isinstance(f, float):
            f = np.ndarray([f])
            squeeze = True
        else:
            squeeze = False

        sgal = np.zeros_like(f)

        if (
            (stochastic_params != () and stochastic_params is not None)
            or (stochastic_kwargs != {} and stochastic_kwargs is not None)
            or stochastic_function is not None
        ):
            if stochastic_function is None:
                stochastic_function = FittedHyperbolicTangentGalacticForeground

            try:
                check = stochastic_function.get_Sh(
                    f, *stochastic_params, **stochastic_kwargs
                )
            except:
                breakpoint()
            sgal[:] = check

        if squeeze:
            sgal = sgal.squeeze()
        return sgal

    @staticmethod
    def stochastic_transform(
        f: float | np.ndarray, Sh: float | np.ndarray, **kwargs: dict
    ) -> float | np.ndarray:
        """Transform from the base stochastic functions to the TDI PSDs.

        **Note**: If not implemented, the transform will return the input.

        Args:
            f: Frequency array.
            Sh: Power spectral density in stochastic term.
            **kwargs: For interoperability.

        Returns:
            Transformed TDI PSD values.

        """
        return Sh


class X1TDISens(Sensitivity):
    @staticmethod
    def transform(
        f: float | np.ndarray,
        Spm: float | np.ndarray,
        Sop: float | np.ndarray,
        **kwargs: dict,
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base sensitivity functions to the XYZ TDI PSDs.\n\n"
            + Sensitivity.transform.__doc__.split("PSDs.\n\n")[-1]
        )

        x = 2.0 * np.pi * lisaLT * f
        return 16.0 * np.sin(x) ** 2 * (2.0 * (1.0 + np.cos(x) ** 2) * Spm + Sop)

    @staticmethod
    def stochastic_transform(
        f: float | np.ndarray, Sh: float | np.ndarray, **kwargs: dict
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base stochastic functions to the XYZ stochastic TDI information.\n\n"
            + Sensitivity.stochastic_transform.__doc__.split("PSDs.\n\n")[-1]
        )
        x = 2.0 * np.pi * lisaLT * f
        t = 4.0 * x**2 * np.sin(x) ** 2
        return Sh * t


class Y1TDISens(X1TDISens):
    __doc__ = X1TDISens.__doc__
    pass


class Z1TDISens(X1TDISens):
    __doc__ = X1TDISens.__doc__
    pass


class XY1TDISens(Sensitivity):
    @staticmethod
    def transform(
        f: float | np.ndarray,
        Spm: float | np.ndarray,
        Sop: float | np.ndarray,
        **kwargs: dict,
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base sensitivity functions to the XYZ TDI PSDs.\n\n"
            + Sensitivity.transform.__doc__.split("PSDs.\n\n")[-1]
        )

        x = 2.0 * np.pi * lisaLT * f
        ## TODO Check the acceleration noise term
        return -4.0 * np.sin(2 * x) * np.sin(x) * (Sop + 4.0 * Spm)

    @staticmethod
    def stochastic_transform(
        f: float | np.ndarray, Sh: float | np.ndarray, **kwargs: dict
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base stochastic functions to the XYZ stochastic TDI information.\n\n"
            + Sensitivity.stochastic_transform.__doc__.split("PSDs.\n\n")[-1]
        )
        x = 2.0 * np.pi * lisaLT * f
        # TODO: check these functions
        # GB = -0.5 of X
        t = -0.5 * (4.0 * x**2 * np.sin(x) ** 2)
        return Sh * t


class ZX1TDISens(XY1TDISens):
    __doc__ = XY1TDISens.__doc__
    pass


class YZ1TDISens(XY1TDISens):
    __doc__ = XY1TDISens.__doc__
    pass


class X2TDISens(Sensitivity):
    @staticmethod
    def transform(
        f: float | np.ndarray,
        Spm: float | np.ndarray,
        Sop: float | np.ndarray,
        **kwargs: dict,
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base sensitivity functions to the XYZ TDI PSDs.\n\n"
            + Sensitivity.transform.__doc__.split("PSDs.\n\n")[-1]
        )

        x = 2.0 * np.pi * lisaLT * f
        ## TODO Check the acceleration noise term
        return (64.0 * np.sin(x) ** 2 * np.sin(2 * x) ** 2 * Sop) + (
            256.0 * (3 + np.cos(2 * x)) * np.cos(x) ** 2 * np.sin(x) ** 4 * Spm
        )

    @staticmethod
    def stochastic_transform(
        f: float | np.ndarray, Sh: float | np.ndarray, **kwargs: dict
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base stochastic functions to the XYZ stochastic TDI information.\n\n"
            + Sensitivity.stochastic_transform.__doc__.split("PSDs.\n\n")[-1]
        )
        x = 2.0 * np.pi * lisaLT * f
        # TODO: check these functions for TDI2
        t = 4.0 * x**2 * np.sin(x) ** 2
        return Sh * t


class Y2TDISens(X2TDISens):
    __doc__ = X2TDISens.__doc__
    pass


class Z2TDISens(X2TDISens):
    __doc__ = X2TDISens.__doc__
    pass


class A1TDISens(Sensitivity):
    @staticmethod
    def transform(
        f: float | np.ndarray,
        Spm: float | np.ndarray,
        Sop: float | np.ndarray,
        **kwargs: dict,
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base sensitivity functions to the A,E TDI PSDs.\n\n"
            + Sensitivity.transform.__doc__.split("PSDs.\n\n")[-1]
        )

        x = 2.0 * np.pi * lisaLT * f
        return (
            8.0
            * np.sin(x) ** 2
            * (
                2.0 * Spm * (3.0 + 2.0 * np.cos(x) + np.cos(2 * x))
                + Sop * (2.0 + np.cos(x))
            )
        )

    @staticmethod
    def stochastic_transform(
        f: float | np.ndarray, Sh: float | np.ndarray, **kwargs: dict
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base stochastic functions to the XYZ stochastic TDI information.\n\n"
            + Sensitivity.stochastic_transform.__doc__.split("PSDs.\n\n")[-1]
        )
        x = 2.0 * np.pi * lisaLT * f
        t = 4.0 * x**2 * np.sin(x) ** 2
        return 1.5 * (Sh * t)


class E1TDISens(A1TDISens):
    __doc__ = A1TDISens.__doc__
    pass


class T1TDISens(Sensitivity):
    @staticmethod
    def transform(
        f: float | np.ndarray,
        Spm: float | np.ndarray,
        Sop: float | np.ndarray,
        **kwargs: dict,
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base sensitivity functions to the T TDI PSDs.\n\n"
            + Sensitivity.transform.__doc__.split("PSDs.\n\n")[-1]
        )

        x = 2.0 * np.pi * lisaLT * f
        return (
            16.0 * Sop * (1.0 - np.cos(x)) * np.sin(x) ** 2
            + 128.0 * Spm * np.sin(x) ** 2 * np.sin(0.5 * x) ** 4
        )

    @staticmethod
    def stochastic_transform(
        f: float | np.ndarray, Sh: float | np.ndarray, **kwargs: dict
    ) -> float | np.ndarray:
        __doc__ = (
            "Transform from the base stochastic functions to the XYZ stochastic TDI information.\n\n"
            + Sensitivity.stochastic_transform.__doc__.split("PSDs.\n\n")[-1]
        )
        x = 2.0 * np.pi * lisaLT * f
        t = 4.0 * x**2 * np.sin(x) ** 2
        return 0.0 * (Sh * t)


class LISASens(Sensitivity):
    @classmethod
    def get_Sn(
        cls,
        f: float | np.ndarray,
        model: Optional[lisa_models.LISAModel | str] = lisa_models.scirdv1,
        average: bool = True,
        **kwargs: dict,
    ) -> float | np.ndarray:
        """Compute the base LISA sensitivity function.

        Args:
            f: Frequency array.
            model: Noise model. Object of type :class:`lisa_models.LISAModel`. It can also be a string corresponding to one of the stock models.
            average: Whether to apply averaging factors to sensitivity curve.
                Antenna response: ``av_resp = np.sqrt(5) if average else 1.0``
                Projection effect: ``Proj = 2.0 / np.sqrt(3) if average else 1.0``
            **kwargs: Keyword arguments to pass to :func:`get_stochastic_contribution`. # TODO: fix

        Returns:
            Sensitivity array.

        """

        # get noise values
        Sa_d, Sop = cls.lisanoises(f, model, unit="displacement")

        all_m = np.sqrt(4.0 * Sa_d + Sop)
        ## Average the antenna response
        av_resp = np.sqrt(5) if average else 1.0

        ## Projection effect
        Proj = 2.0 / np.sqrt(3) if average else 1.0

        ## Approximative transfer function
        f0 = 1.0 / (2.0 * lisaLT)
        a = 0.41
        T = np.sqrt(1 + (f / (a * f0)) ** 2)
        sens = (av_resp * Proj * T * all_m / lisaL) ** 2

        # will add zero if ignored
        sens += cls.get_stochastic_contribution(f, **kwargs)
        return sens


class CornishLISASens(LISASens):
    """PSD from https://arxiv.org/pdf/1803.01944.pdf

    Power Spectral Density for the LISA detector assuming it has been active for a year.
    I found an analytic version in one of Niel Cornish's paper which he submitted to the arXiv in
    2018. I evaluate the PSD at the frequency bins found in the signal FFT.

    PSD obtained from: https://arxiv.org/pdf/1803.01944.pdf

    """

    @staticmethod
    def get_Sn(
        f: float | np.ndarray, average: bool = True, **kwargs: dict
    ) -> float | np.ndarray:
        # TODO: documentation here

        sky_averaging_constant = 20.0 / 3.0 if average else 1.0

        L = 2.5 * 10**9  # Length of LISA arm
        f0 = 19.09 * 10 ** (-3)  # transfer frequency

        # Optical Metrology Sensor
        Poms = ((1.5e-11) * (1.5e-11)) * (1 + np.power((2e-3) / f, 4))

        # Acceleration Noise
        Pacc = (
            (3e-15)
            * (3e-15)
            * (1 + (4e-4 / f) * (4e-4 / f))
            * (1 + np.power(f / (8e-3), 4))
        )

        # constants for Galactic background after 1 year of observation
        alpha = 0.171
        beta = 292
        k = 1020
        gamma = 1680
        f_k = 0.00215

        # Galactic background contribution
        Sc = (
            9e-45
            * np.power(f, -7 / 3)
            * np.exp(-np.power(f, alpha) + beta * f * np.sin(k * f))
            * (1 + np.tanh(gamma * (f_k - f)))
        )

        # PSD
        PSD = (sky_averaging_constant) * (
            (10 / (3 * L * L))
            * (Poms + (4 * Pacc) / (np.power(2 * np.pi * f, 4)))
            * (1 + 0.6 * (f / f0) * (f / f0))
            + Sc
        )

        return PSD


class FlatPSDFunction(LISASens):
    """White Noise PSD function."""

    @staticmethod
    def get_Sn(f: float | np.ndarray, val: float, **kwargs: dict) -> float | np.ndarray:
        # TODO: documentation here
        out = np.full_like(f, val)
        if isinstance(f, float):
            out = out.item()
        return out


class SensitivityMatrix:
    """Container to hold sensitivity information.

    Args:
        f: Frequency array.
        sens_mat: Input sensitivity list. The shape of the nested lists should represent the shape of the
            desired matrix. Each entry in the list must be an array, :class:`Sensitivity`-derived object,
            or a string corresponding to the :class:`Sensitivity` object.
        **sens_kwargs: Keyword arguments to pass to :method:`Sensitivity.get_Sn`.

    """

    def __init__(
        self,
        f: np.ndarray,
        sens_mat: List[List[np.ndarray | Sensitivity]]
        | List[np.ndarray | Sensitivity]
        | np.ndarray
        | Sensitivity,
        *sens_args: tuple,
        **sens_kwargs: dict,
    ) -> None:
        self.frequency_arr = f
        self.data_length = len(self.frequency_arr)
        self.sens_args = sens_args
        self.sens_kwargs = sens_kwargs
        self.sens_mat = sens_mat

    @property
    def frequency_arr(self) -> np.ndarray:
        return self._frequency_arr

    @frequency_arr.setter
    def frequency_arr(self, frequency_arr: np.ndarray) -> None:
        assert frequency_arr.dtype == np.float64 or frequency_arr.dtype == float
        assert frequency_arr.ndim == 1
        self._frequency_arr = frequency_arr

    def update_frequency_arr(self, frequency_arr: np.ndarray) -> None:
        """Update class with new frequency array.

        Args:
            frequency_arr: Frequency array.

        """
        self.frequency_arr = frequency_arr
        self.sens_mat = self.sens_mat_input

    def update_model(self, model: lisa_models.LISAModel) -> None:
        """Update class with new sensitivity model.

        Args:
            model: Noise model. Object of type :class:`lisa_models.LISAModel`. It can also be a string corresponding to one of the stock models.

        """
        self.sens_kwargs["model"] = model
        self.sens_mat = self.sens_mat_input

    def update_stochastic(self, **kwargs: dict) -> None:
        """Update class with new stochastic function.

        Args:
            **kwargs: Keyword arguments update for :func:`lisatools.sensitivity.Sensitivity.get_stochastic_contribution`.
                This operation will combine the new and old kwarg dictionaries, updating any
                old information with any added corresponding new information. **Note**: any old information
                that is not updated will remain in place.

        """
        self.sens_kwargs = {**self.sens_kwargs, **kwargs}
        self.sens_mat = self.sens_mat_input

    @property
    def sens_mat(self) -> np.ndarray:
        """Get sensitivity matrix."""
        return self._sens_mat

    @sens_mat.setter
    def sens_mat(
        self,
        sens_mat: List[List[np.ndarray | Sensitivity]]
        | List[np.ndarray | Sensitivity]
        | np.ndarray
        | Sensitivity,
    ) -> None:
        """Set sensitivity matrix."""
        self.sens_mat_input = deepcopy(sens_mat)
        self._sens_mat = np.asarray(sens_mat, dtype=object)

        # not an
        new_out = np.full(len(self._sens_mat.flatten()), None, dtype=object)
        self.return_shape = self._sens_mat.shape
        for i in range(len(self._sens_mat.flatten())):
            current_sens = self._sens_mat.flatten()[i]
            if hasattr(current_sens, "get_Sn") or isinstance(current_sens, str):
                new_out[i] = get_sensitivity(
                    self.frequency_arr,
                    *self.sens_args,
                    sens_fn=current_sens,
                    **self.sens_kwargs,
                )

            elif isinstance(current_sens, np.ndarray) or isinstance(
                current_sens, cp.ndarray
            ):
                new_out[i] = current_sens
            else:
                raise ValueError

        # setup in array form
        self._sens_mat = np.asarray(list(new_out), dtype=float).reshape(
            self.return_shape + (-1,)
        )

    def __getitem__(self, index: Any) -> np.ndarray:
        """Indexing the class indexes the array."""
        return self.sens_mat[index]

    @property
    def ndim(self) -> int:
        """Dimensionality of sens mat array."""
        return self.sens_mat.ndim

    def flatten(self) -> np.ndarray:
        """Flatten sens mat array."""
        return self.sens_mat.reshape(-1, self.sens_mat.shape[-1])

    @property
    def shape(self) -> tuple:
        """Shape of sens mat array."""
        return self.sens_mat.shape

    def loglog(
        self,
        ax: Optional[plt.Axes] = None,
        fig: Optional[plt.Figure] = None,
        inds: Optional[int | tuple] = None,
        char_strain: Optional[bool] = False,
        **kwargs: dict,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Produce a log-log plot of the sensitivity.

        Args:
            ax: Matplotlib Axes objects to add plots. Either a list of Axes objects or a single Axes object.
            fig: Matplotlib figure object.
            inds: Integer index to select out which data to add to a single access.
                A list can be provided if ax is a list. They must be the same length.
            char_strain: If ``True``, plot in characteristic strain representation. **Note**: assumes the sensitivity
                is input as power spectral density.
            **kwargs: Keyword arguments to be passed to ``loglog`` function in matplotlib.

        Returns:
            Matplotlib figure and axes objects in a 2-tuple.


        """
        if (ax is None and fig is None) or (
            ax is not None and (isinstance(ax, list) or isinstance(ax, np.ndarray))
        ):
            if ax is None and fig is None:
                outer_shape = self.shape[:-1]
                if len(outer_shape) == 2:
                    nrows = outer_shape[0]
                    ncols = outer_shape[1]
                elif len(outer_shape) == 1:
                    nrows = 1
                    ncols = outer_shape[0]

                fig, ax = plt.subplots(nrows, ncols, sharex=True, sharey=True)
                try:
                    ax = ax.ravel()
                except AttributeError:
                    ax = [ax]  # just one axis object, no list

            else:
                assert len(ax) == np.prod(self.shape[:-1])

            for i in range(np.prod(self.shape[:-1])):
                plot_in = self.flatten()[i]
                if char_strain:
                    plot_in = np.sqrt(self.frequency_arr * plot_in)
                ax[i].loglog(self.frequency_arr, plot_in, **kwargs)

        elif fig is not None:
            raise NotImplementedError

        elif isinstance(ax, plt.axes):
            if inds is None:
                raise ValueError(
                    "When passing a single axes object for `ax`, but also pass `inds` kwarg."
                )
            plot_in = self.sens_mat[inds]
            if char_strain:
                plot_in = np.sqrt(self.frequency_arr * plot_in)
            ax.loglog(self.frequency_arr, plot_in, **kwargs)

        else:
            raise ValueError(
                "ax must be a list of axes objects or a single axes object."
            )

        return (fig, ax)


class XYZ1SensitivityMatrix(SensitivityMatrix):
    """Default sensitivity matrix for XYZ (TDI 1)

    This is 3x3 symmetric matrix.

    Args:
        f: Frequency array.
        **sens_kwargs: Keyword arguments to pass to :method:`Sensitivity.get_Sn`.

    """

    def __init__(self, f: np.ndarray, **sens_kwargs: dict) -> None:
        sens_mat = [
            [X1TDISens, XY1TDISens, ZX1TDISens],
            [XY1TDISens, Y1TDISens, YZ1TDISens],
            [ZX1TDISens, YZ1TDISens, Z1TDISens],
        ]
        super().__init__(f, sens_mat, **sens_kwargs)


class AET1SensitivityMatrix(SensitivityMatrix):
    """Default sensitivity matrix for AET (TDI 1)

    This is just an array because no cross-terms.

    Args:
        f: Frequency array.
        **sens_kwargs: Keyword arguments to pass to :method:`Sensitivity.get_Sn`.

    """

    def __init__(self, f: np.ndarray, **sens_kwargs: dict) -> None:
        sens_mat = [A1TDISens, E1TDISens, T1TDISens]
        super().__init__(f, sens_mat, **sens_kwargs)


class AE1SensitivityMatrix(SensitivityMatrix):
    """Default sensitivity matrix for AE (no T) (TDI 1)

    Args:
        f: Frequency array.
        **sens_kwargs: Keyword arguments to pass to :method:`Sensitivity.get_Sn`.

    """

    def __init__(self, f: np.ndarray, **sens_kwargs: dict) -> None:
        sens_mat = [A1TDISens, E1TDISens]
        super().__init__(f, sens_mat, **sens_kwargs)


class LISASensSensitivityMatrix(SensitivityMatrix):
    """Default sensitivity matrix adding :class:`LISASens` for the specified number of channels.

    Args:
        f: Frequency array.
        nchannels: Number of channels.
        **sens_kwargs: Keyword arguments to pass to :method:`Sensitivity.get_Sn`.

    """

    def __init__(self, f: np.ndarray, nchannels: int, **sens_kwargs: dict) -> None:
        sens_mat = [LISASens for _ in range(nchannels)]
        super().__init__(f, sens_mat, **sens_kwargs)


def get_sensitivity(
    f: float | np.ndarray,
    *args: tuple,
    sens_fn: Optional[Sensitivity | str] = LISASens,
    return_type="PSD",
    **kwargs,
) -> float | np.ndarray:
    """Generic sensitivity generator

    Same interface to many sensitivity curves.

    Args:
        f: Frequency array.
        *args: Any additional arguments for the sensitivity function ``get_Sn`` method.
        sens_fn: String or class that represents the name of the desired PSD function.
        return_type: Described the desired output. Choices are ASD,
            PSD, or char_strain (characteristic strain). Default is ASD.
        **kwargs: Keyword arguments to pass to sensitivity function ``get_Sn`` method.

    Return:
        Sensitivity values.

    """

    if isinstance(sens_fn, str):
        try:
            sensitivity = globals()[sens_fn]
        except KeyError:
            raise ValueError(
                f"{sens_fn} sensitivity is not available. Available stock sensitivities are"
            )

    elif hasattr(sens_fn, "get_Sn"):
        sensitivity = sens_fn

    else:
        raise ValueError(
            "sens_fn must be a string for a stock option or a class with a get_Sn method."
        )

    PSD = sensitivity.get_Sn(f, *args, **kwargs)

    if return_type == "PSD":
        return PSD

    elif return_type == "ASD":
        return PSD ** (1 / 2)

    elif return_type == "char_strain":
        return (f * PSD) ** (1 / 2)

    else:
        raise ValueError("return_type must be PSD, ASD, or char_strain.")


__stock_sens_options__ = [
    "X1TDISens",
    "Y1TDISens",
    "Z1TDISens",
    "XY1TDISens",
    "YZ1TDISens",
    "ZX1TDISens",
    "A1TDISens",
    "E1TDISens",
    "T1TDISens",
    "X2TDISens",
    "Y2TDISens",
    "Z2TDISens",
    "LISASens",
    "CornishLISASens",
    "FlatPSDFunction",
]


def get_stock_sensitivity_options() -> List[Sensitivity]:
    """Get stock options for sensitivity curves.

    Returns:
        List of stock sensitivity options.

    """
    return __stock_sens_options__


__stock_sensitivity_mat_options__ = [
    "XYZ1SensitivityMatrix",
    "AET1SensitivityMatrix",
    "AE1SensitivityMatrix",
]


def get_stock_sensitivity_matrix_options() -> List[SensitivityMatrix]:
    """Get stock options for sensitivity matrix.

    Returns:
        List of stock sensitivity matrix options.

    """
    return __stock_sensitivity_mat_options__



import cupy as xp
import numpy as np
from lisatools.sensitivity import get_sensitivity
import time


class MultiGPUDataHolder:
    def __init__(self, gpus, channel1_data, channel2_data, channel1_base_data, channel2_base_data, channel1_psd, channel2_psd, channel1_lisasens, channel2_lisasens, df, base_injections=None, base_psd=None):

        if isinstance(gpus, int):
            gpus = [gpus]

        self.df = df
        
        if not isinstance(gpus, list) or not isinstance(gpus[0], int):
            raise ValueError("gpus must be an integer or a list of integers.")

        self.gpus = gpus
        self.num_gpus = len(gpus)
        # need to be numpy coming in to now make memory large
        self.ntemps, self.nwalkers, self.data_length = channel1_data.shape
        self.total_number = self.nwalkers
        self.walker_indices = np.arange(self.nwalkers)
        self.overall_indices_flat = np.arange(2 * self.nwalkers)  # evens and odds

        self.fd = np.arange(self.data_length) * df

        self.base_injections = base_injections
        self.base_psd = base_psd

        self.map = self.overall_indices_flat.copy()

        num_per_split = self.total_number // self.num_gpus + 1 * (self.total_number % self.num_gpus != 0)
        # gpu arangement
        self.gpu_split_inds = np.arange(num_per_split, self.total_number, num_per_split)

        self.gpu_splits = [
            np.split(self.overall_indices_flat[:self.nwalkers], self.gpu_split_inds),
            np.split(self.overall_indices_flat[self.nwalkers:], self.gpu_split_inds)
        ]
        self.gpu_splits = [np.concatenate([self.gpu_splits[0][i], self.gpu_splits[1][i]]) for i in range(len(self.gpu_splits[0]))]

        self.gpus_for_each_data = [np.full_like(gpu_split, gpu) for gpu_split, gpu in zip( self.gpu_splits, self.gpus)]
        self.mempool = xp.get_default_memory_pool()

        self.channel1_data = [None for _ in range(self.num_gpus)]
        self.channel2_data = [None for _ in range(self.num_gpus)]
        self.channel1_base_data = [None for _ in range(self.num_gpus)]
        self.channel2_base_data = [None for _ in range(self.num_gpus)]
        self.channel1_psd = [None for _ in range(self.num_gpus)]
        self.channel2_psd = [None for _ in range(self.num_gpus)]
        self.channel1_lisasens = [None for _ in range(self.num_gpus)]
        self.channel2_lisasens = [None for _ in range(self.num_gpus)]
        return_to_main = xp.cuda.runtime.getDevice()
        for gpu_i, (gpu, gpu_split_tmp) in enumerate(zip(self.gpus, self.gpu_splits)):
            gpu_split = gpu_split_tmp[gpu_split_tmp < self.nwalkers]
            walker_inds_gpu_here = self.walker_indices[gpu_split]

            with xp.cuda.device.Device(gpu):
                
                self.channel1_data[gpu_i] = xp.zeros(2 * walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel1_data.dtype)
                self.channel2_data[gpu_i] = xp.zeros(2 * walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel1_data.dtype)
                self.channel1_base_data[gpu_i] = xp.zeros(walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel1_data.dtype)
                self.channel2_base_data[gpu_i] = xp.zeros(walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel1_data.dtype)
                self.channel1_psd[gpu_i] = xp.zeros(walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel1_psd.dtype)
                self.channel2_psd[gpu_i] = xp.zeros(walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel2_psd.dtype)
                self.channel1_lisasens[gpu_i] = xp.zeros(walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel1_lisasens.dtype)
                self.channel2_lisasens[gpu_i] = xp.zeros(walker_inds_gpu_here.shape[0] * channel1_data.shape[-1], dtype=channel2_lisasens.dtype)
                
                for data_i, walker_ind in enumerate(walker_inds_gpu_here):
                    inds_slice = slice(data_i * channel1_data.shape[-1], (data_i + 1) * channel1_data.shape[-1])
                    inds_slice_even = slice(data_i * channel1_data.shape[-1], (data_i + 1) * channel1_data.shape[-1])
                    inds_slice_odd = slice((self.nwalkers + data_i) * channel1_data.shape[-1], (self.nwalkers + data_i + 1) * channel1_data.shape[-1])
                    
                    tmp_data1 = xp.asarray(channel1_data[0, walker_ind])
                    self.channel1_data[gpu_i][inds_slice_even] = tmp_data1
                    del tmp_data1
                    self.mempool.free_all_blocks()

                    tmp_data1 = xp.asarray(channel1_data[1, walker_ind])
                    self.channel1_data[gpu_i][inds_slice_odd] = tmp_data1
                    del tmp_data1
                    self.mempool.free_all_blocks()

                    tmp_data2 = xp.asarray(channel2_data[0, walker_ind])
                    self.channel2_data[gpu_i][inds_slice_even] = tmp_data2
                    del tmp_data2
                    self.mempool.free_all_blocks()

                    tmp_data2 = xp.asarray(channel2_data[1, walker_ind])
                    self.channel2_data[gpu_i][inds_slice_odd] = tmp_data2
                    del tmp_data2
                    self.mempool.free_all_blocks()

                    # TODO: reconsider use of this data since it is just for checking LL
                    tmp_base_data1 = xp.asarray(channel1_base_data[0, walker_ind])
                    self.channel1_base_data[gpu_i][inds_slice_even] = tmp_base_data1
                    del tmp_base_data1
                    self.mempool.free_all_blocks()

                    tmp_base_data2 = xp.asarray(channel2_base_data[0, walker_ind])
                    self.channel2_base_data[gpu_i][inds_slice_even] = tmp_base_data2
                    del tmp_base_data2
                    self.mempool.free_all_blocks()

                    tmp_psd1 = xp.asarray(channel1_psd[0, walker_ind])
                    self.channel1_psd[gpu_i][inds_slice] = tmp_psd1
                    del tmp_psd1
                    self.mempool.free_all_blocks()

                    tmp_psd2 = xp.asarray(channel2_psd[0, walker_ind])
                    self.channel2_psd[gpu_i][inds_slice] = tmp_psd2
                    del tmp_psd2
                    self.mempool.free_all_blocks()

                    tmp_lisasens1 = xp.asarray(channel1_lisasens[0, walker_ind])
                    self.channel1_lisasens[gpu_i][inds_slice] = tmp_lisasens1
                    del tmp_lisasens1
                    self.mempool.free_all_blocks()

                    tmp_lisasens2 = xp.asarray(channel2_lisasens[0, walker_ind])
                    self.channel2_lisasens[gpu_i][inds_slice] = tmp_lisasens2
                    del tmp_lisasens2
                    self.mempool.free_all_blocks()
        
        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

    def reshape_list(self, input_value):
        return [
            self.reshape(tmp) for tmp in input_value
        ]

    def reshape(self, input_value):
        return input_value.reshape(-1, self.data_length)

    @property
    def data_list(self):
        return [self.channel1_data, self.channel2_data]

    @property
    def base_data_list(self):
        return [self.channel1_base_data, self.channel2_base_data]

    @property
    def psd_list(self):
        return [self.channel1_psd, self.channel2_psd]

    @property
    def lisasens_list(self):
        return [self.channel1_lisasens, self.channel2_lisasens]

    @property
    def data_shaped(self):
        tmp1 = [self.channel1_data[i][:self.nwalkers * self.data_length] + self.channel1_data[i][self.nwalkers * self.data_length:] - self.channel1_base_data[i][:] for i in range(len(self.channel1_data))]
        tmp2 = [self.channel2_data[i][:self.nwalkers * self.data_length] + self.channel2_data[i][self.nwalkers * self.data_length:] - self.channel2_base_data[i][:] for i in range(len(self.channel2_data))]
            
        return [
            self.reshape_list(tmp1),
            self.reshape_list(tmp2),
        ]

    @property
    def data_shaped_2_parts(self):
        return [
            self.reshape_list(self.channel1_data),
            self.reshape_list(self.channel2_data),
        ]

    @property
    def data_shaped_base(self):
        return [
            self.reshape_list(self.channel1_base_data),
            self.reshape_list(self.channel2_base_data),
        ]

    @property
    def psd_shaped(self):
        return [
            self.reshape_list(self.channel1_psd),
            self.reshape_list(self.channel2_psd),
        ]

    @property
    def lisasens_shaped(self):
        return [
            self.reshape_list(self.channel1_lisasens),
            self.reshape_list(self.channel2_lisasens),
        ]

    @property
    def map(self):
        return self._map

    @map.setter
    def map(self, map):
        if not isinstance(map, np.ndarray) or len(map) != 2 * self.total_number or map.dtype != np.int64:
            raise ValueError("map input must be a numpy array of np.int64 that is the same length as the number of gpu holder slots.")
        self._map = map

    @property
    def full_length(self):
        return self.ntemps * self.nwalkers * self.data_length

    def get_mapped_indices(self, inds_in):
        if (not isinstance(inds_in, np.ndarray) and not isinstance(inds_in, xp.ndarray)) or ((inds_in.dtype != np.int64 and inds_in.dtype != xp.int32)):
            raise ValueError("inds_in input must be a numpy array of np.int64.")

        if isinstance(inds_in, np.ndarray):
            xp_here = np
        else:
            xp_here = xp
        return xp_here.asarray(self.map)[inds_in]

    def set_psd_from_arrays(self, A_vals_in, E_vals_in, overall_inds=None):

        if overall_inds is None:
            overall_inds = np.arange(self.ntemps * self.nwalkers)

        assert len(A_vals_in) == len(E_vals_in) == len(overall_inds)
        return_to_main = xp.cuda.runtime.getDevice()

        fd_gpu = [None for _ in self.gpus]
        A_tmp = [None for _ in self.gpus]
        E_tmp = [None for _ in self.gpus]
        # st = time.perf_counter()
        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                fd_gpu[gpu_i] = xp.asarray(self.fd)
                for i, (overall_index) in enumerate(overall_inds):
                    
                    if overall_index not in gpu_split:
                        continue

                    overall_index_here = overall_index - gpu_split.min().item()
                    
                    A_tmp[gpu_i] = xp.asarray(A_vals_in[i])
                    A_tmp[gpu_i][0] = A_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel1_psd[gpu_i][inds_slice] = A_tmp[gpu_i]
                    if xp.any(A_tmp[gpu_i] < 0.0):
                        breakpoint()

                    E_tmp[gpu_i] = xp.asarray(E_vals_in[i])
                    E_tmp[gpu_i][0] = E_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel2_psd[gpu_i][inds_slice] = E_tmp[gpu_i]
                    if xp.any(E_tmp[gpu_i] < 0.0):
                        breakpoint()

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                del fd_gpu[gpu_i], A_tmp[gpu_i], E_tmp[gpu_i]
                xp.get_default_memory_pool().free_all_blocks()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

    def set_lisasens_from_arrays(self, A_vals_in, E_vals_in, overall_inds=None):

        if overall_inds is None:
            overall_inds = np.arange(self.ntemps * self.nwalkers)

        assert len(A_vals_in) == len(E_vals_in) == len(overall_inds)
        return_to_main = xp.cuda.runtime.getDevice()

        fd_gpu = [None for _ in self.gpus]
        A_tmp = [None for _ in self.gpus]
        E_tmp = [None for _ in self.gpus]
        # st = time.perf_counter()
        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                fd_gpu[gpu_i] = xp.asarray(self.fd)
                for i, (overall_index) in enumerate(overall_inds):
                    
                    if overall_index not in gpu_split:
                        continue

                    overall_index_here = overall_index - gpu_split.min().item()
                    
                    A_tmp[gpu_i] = xp.asarray(A_vals_in[i])
                    A_tmp[gpu_i][0] = A_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel1_lisasens[gpu_i][inds_slice] = A_tmp[gpu_i]
                    if xp.any(A_tmp[gpu_i] < 0.0):
                        breakpoint()

                    E_tmp[gpu_i] = xp.asarray(E_vals_in[i])
                    E_tmp[gpu_i][0] = E_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel2_lisasens[gpu_i][inds_slice] = E_tmp[gpu_i]
                    if xp.any(E_tmp[gpu_i] < 0.0):
                        breakpoint()

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                del fd_gpu[gpu_i], A_tmp[gpu_i], E_tmp[gpu_i]
                xp.get_default_memory_pool().free_all_blocks()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

    def add_templates_from_arrays_to_residuals(self, A_vals_in, E_vals_in, overall_inds=None):

        if overall_inds is None:
            overall_inds = np.arange(self.ntemps * self.nwalkers)

        assert len(A_vals_in) == len(E_vals_in) == len(overall_inds)
        return_to_main = xp.cuda.runtime.getDevice()

        fd_gpu = [None for _ in self.gpus]
        A_tmp = [None for _ in self.gpus]
        E_tmp = [None for _ in self.gpus]
        # st = time.perf_counter()
        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                for i, (overall_index) in enumerate(overall_inds):
                    
                    if overall_index not in gpu_split:
                        continue

                    overall_index_here = overall_index - gpu_split.min().item()
                    
                    A_tmp[gpu_i] = xp.asarray(A_vals_in[i])
                    A_tmp[gpu_i][0] = A_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel1_data[gpu_i][inds_slice] -= A_tmp[gpu_i]
                    if xp.any(xp.isnan(A_tmp[gpu_i] < 0.0)):
                        breakpoint()

                    E_tmp[gpu_i] = xp.asarray(E_vals_in[i])
                    E_tmp[gpu_i][0] = E_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel2_data[gpu_i][inds_slice] -= E_tmp[gpu_i]
                    if xp.any(xp.isnan(E_tmp[gpu_i] < 0.0)):
                        breakpoint()

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                del A_tmp[gpu_i], E_tmp[gpu_i]
                xp.get_default_memory_pool().free_all_blocks()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()


    def set_psd_vals(self, psd_params, overall_inds=None, foreground_params=None):

        if overall_inds is None:
            overall_inds = np.arange(self.ntemps * self.nwalkers)
        return_to_main = xp.cuda.runtime.getDevice()

        fd_gpu = [None for _ in self.gpus]
        A_tmp = [None for _ in self.gpus]
        E_tmp = [None for _ in self.gpus]
        # st = time.perf_counter()
        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                fd_gpu[gpu_i] = xp.asarray(self.fd)
                for i, (overall_index) in enumerate(overall_inds):
                    
                    if overall_index not in gpu_split:
                        continue

                    overall_index_here = overall_index - gpu_split.min().item()
                    
                    if foreground_params is not None:
                        foreground_pars_in = foreground_params[i]
                    else:
                        foreground_pars_in = None

                    psd_params_A_in = psd_params[i][:2]
                    
                    A_tmp[gpu_i] = get_sensitivity(fd_gpu[gpu_i], sens_fn="noisepsd_AE", model=psd_params_A_in, foreground_params=foreground_pars_in, xp=xp)
                    A_tmp[gpu_i][0] = A_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel1_psd[gpu_i][inds_slice] = A_tmp[gpu_i]
                    if xp.any(A_tmp[gpu_i] < 0.0):
                        breakpoint()

                    psd_params_E_in = psd_params[i][2:]

                    E_tmp[gpu_i] = get_sensitivity(fd_gpu[gpu_i], sens_fn="noisepsd_AE", model=psd_params_E_in, foreground_params=foreground_pars_in, xp=xp)
                    E_tmp[gpu_i][0] = E_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel2_psd[gpu_i][inds_slice] = E_tmp[gpu_i]
                    if xp.any(E_tmp[gpu_i] < 0.0):
                        breakpoint()

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                del fd_gpu[gpu_i], A_tmp[gpu_i], E_tmp[gpu_i]
                xp.get_default_memory_pool().free_all_blocks()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

        # et = time.perf_counter()
        # print("fill", et - st)  

    def set_lisasens_vals(self, lisasens_params, overall_inds=None, foreground_params=None):

        if overall_inds is None:
            overall_inds = np.arange(self.ntemps * self.nwalkers)
        return_to_main = xp.cuda.runtime.getDevice()

        fd_gpu = [None for _ in self.gpus]
        A_tmp = [None for _ in self.gpus]
        E_tmp = [None for _ in self.gpus]
        # st = time.perf_counter()
        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                fd_gpu[gpu_i] = xp.asarray(self.fd)
                for i, (overall_index) in enumerate(overall_inds):
                    
                    if overall_index not in gpu_split:
                        continue

                    overall_index_here = overall_index - gpu_split.min().item()
                    
                    if foreground_params is not None:
                        foreground_pars_in = foreground_params[i]
                    else:
                        foreground_pars_in = None

                    lisasens_params_A_in = lisasens_params[i][:2]
                    
                    A_tmp[gpu_i] = get_sensitivity(fd_gpu[gpu_i], sens_fn="lisasens", model=lisasens_params_A_in, foreground_params=foreground_pars_in, xp=xp)
                    A_tmp[gpu_i][0] = A_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel1_lisasens[gpu_i][inds_slice] = A_tmp[gpu_i]
                    if xp.any(A_tmp[gpu_i] < 0.0):
                        breakpoint()

                    lisasens_params_E_in = lisasens_params[i][2:]

                    E_tmp[gpu_i] = get_sensitivity(fd_gpu[gpu_i], sens_fn="lisasens", model=lisasens_params_E_in, foreground_params=foreground_pars_in, xp=xp)
                    E_tmp[gpu_i][0] = E_tmp[gpu_i][1]
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    self.channel2_lisasens[gpu_i][inds_slice] = E_tmp[gpu_i]
                    if xp.any(E_tmp[gpu_i] < 0.0):
                        breakpoint()

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                del fd_gpu[gpu_i], A_tmp[gpu_i], E_tmp[gpu_i]
                xp.get_default_memory_pool().free_all_blocks()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

        # et = time.perf_counter()
        # print("fill", et - st)  

    def get_psd_term(self, overall_inds=None):

        reshape = False
        if overall_inds is None:
            reshape = True
            overall_inds = np.arange(self.nwalkers)
                
        return_to_main = xp.cuda.runtime.getDevice()

        psd_term = np.zeros_like(overall_inds, dtype=float)

        # st = time.perf_counter()
        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()
                for i, (overall_index) in enumerate(overall_inds):
                    if overall_index not in gpu_split:
                        continue

                    overall_index_here = overall_index - gpu_split.min().item()
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)

                    psd_term_here = xp.sum((xp.log(self.channel1_psd[gpu_i][inds_slice]) + xp.log(self.channel2_psd[gpu_i][inds_slice]))).get().item()
                    xp.cuda.runtime.deviceSynchronize()
                    if np.isnan(psd_term_here):
                        breakpoint()
                    psd_term[i] = psd_term_here
                    

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                xp.get_default_memory_pool().free_all_blocks()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

        # if reshape:
        #     psd_term = psd_term.reshape(self.ntemps, self.nwalkers)
            
        # et = time.perf_counter()
        # print("get psd term", et - st)
        return psd_term

    def sub_in_data_and_psd(self, data, psd, lisasens):
        """Must be the same size at current data
        
        
        """
        assert len(self.gpus) == 1
        gpu_i = 0

        # adjust psd
        self.channel1_psd[gpu_i][:] = xp.asarray(psd[0].flatten())
        self.channel2_psd[gpu_i][:] = xp.asarray(psd[1].flatten())

        # adjust lisasens
        self.channel1_lisasens[gpu_i][:] = xp.asarray(lisasens[0].flatten())
        self.channel2_lisasens[gpu_i][:] = xp.asarray(lisasens[1].flatten())

        # remove injected data + previous templates
        self.channel1_data[gpu_i][:self.nwalkers * self.data_length] -= self.channel1_base_data[gpu_i][:]
        self.channel1_data[gpu_i][self.nwalkers * self.data_length:] -= self.channel1_base_data[gpu_i][:]

        self.channel2_data[gpu_i][:self.nwalkers * self.data_length] -= self.channel2_base_data[gpu_i][:]
        self.channel2_data[gpu_i][self.nwalkers * self.data_length:] -= self.channel2_base_data[gpu_i][:]

        # change injected data + other templates in base
        self.channel1_base_data[gpu_i][:] = xp.asarray(data[0].flatten())
        self.channel2_base_data[gpu_i][:] = xp.asarray(data[1].flatten())

        # re-add to channel data
        self.channel1_data[gpu_i][:self.nwalkers * self.data_length] += self.channel1_base_data[gpu_i][:]
        self.channel1_data[gpu_i][self.nwalkers * self.data_length:] += self.channel1_base_data[gpu_i][:]

        self.channel2_data[gpu_i][:self.nwalkers * self.data_length] += self.channel2_base_data[gpu_i][:]
        self.channel2_data[gpu_i][self.nwalkers * self.data_length:] += self.channel2_base_data[gpu_i][:]

        return        


    def get_inner_product(self, *args, overall_inds=None, band_edge_inds=None, **kwargs):
        reshape = False
        if overall_inds is None:
            reshape = True
            overall_inds = np.arange(self.nwalkers)
                
        return_to_main = xp.cuda.runtime.getDevice()

        if band_edge_inds is None:
            inner_term = np.zeros_like(overall_inds, dtype=float)
        else:
            inner_term = np.zeros((overall_inds.shape[0], band_edge_inds.shape[0] - 1), dtype=float)

        data_tmp1 = [None for _ in self.gpus]
        data_tmp2 = [None for _ in self.gpus]
        psd_tmp1 = [None for _ in self.gpus]
        psd_tmp2 = [None for _ in self.gpus]

        # st = time.perf_counter()
        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()
                for i, (overall_index) in enumerate(overall_inds):
                    if overall_index not in gpu_split:
                        continue

                    overall_index_here = overall_index - gpu_split.min().item()
                    inds_slice = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    inds_slice_even = slice(overall_index_here * self.data_length, (overall_index_here + 1) * self.data_length)
                    inds_slice_odd = slice((self.nwalkers + overall_index_here) * self.data_length, (self.nwalkers + overall_index_here + 1) * self.data_length)

                    data_tmp1[gpu_i] = self.channel1_data[gpu_i][inds_slice_even] + self.channel1_data[gpu_i][inds_slice_odd] - self.channel1_base_data[gpu_i][inds_slice]
                    psd_tmp1[gpu_i] = self.channel1_psd[gpu_i][inds_slice]
                    data_tmp2[gpu_i] = self.channel2_data[gpu_i][inds_slice_even] + self.channel2_data[gpu_i][inds_slice_odd] - self.channel2_base_data[gpu_i][inds_slice]
                    psd_tmp2[gpu_i] = self.channel2_psd[gpu_i][inds_slice]

                    if band_edge_inds is None:
                        inner_here = self.df * 4 * xp.sum(
                                data_tmp1[gpu_i].conj() * data_tmp1[gpu_i] / psd_tmp1[gpu_i]
                                + data_tmp2[gpu_i].conj() * data_tmp2[gpu_i] / psd_tmp2[gpu_i],
                        ).real.item()

                    else:
                        inner_here_tmp = self.df * 4 * xp.cumsum(
                                data_tmp1[gpu_i].conj() * data_tmp1[gpu_i] / psd_tmp1[gpu_i]
                                + data_tmp2[gpu_i].conj() * data_tmp2[gpu_i] / psd_tmp2[gpu_i],
                        ).real[band_edge_inds]
                        inner_here_tmp[1:] -= inner_here_tmp[:-1]
                        inner_here = inner_here_tmp[1:]

                    # if overall_index_here == 11:
                    #     # for w in range(3951, 3951 + 420, 25):
                    #     #     print(f"INCHECKIT : {w} {data_tmp1[gpu_i][w].real} {data_tmp1[gpu_i][w].imag}, {self.channel1_data[gpu_i][inds_slice_even][w].real} {self.channel1_data[gpu_i][inds_slice_even][w].imag}, {self.channel1_data[gpu_i][inds_slice_odd][w].real} {self.channel1_data[gpu_i][inds_slice_odd][w].imag}, {self.channel1_base_data[gpu_i][inds_slice][w].real} {self.channel1_base_data[gpu_i][inds_slice][w].imag}")
                        
                    #     inner_here_check = self.df * 4 * xp.cumsum(
                    #             data_tmp1[gpu_i][3951:3951 + 420].conj() * data_tmp1[gpu_i][3951:3951 + 420] / psd_tmp1[gpu_i][3951:3951 + 420]
                    #             + data_tmp2[gpu_i][3951:3951 + 420].conj() * data_tmp2[gpu_i][3951:3951 + 420] / psd_tmp2[gpu_i][3951:3951 + 420],
                    #     ).real
                    #     # print("INSIDE INNER: ", -1/2 * inner_here_check)  # , data_tmp1[gpu_i][3811], self.channel1_data[gpu_i][inds_slice_even][3811], self.channel1_data[gpu_i][inds_slice_odd][3811], self.channel1_base_data[gpu_i][inds_slice][3811])
                    #     if "stop" in kwargs and kwargs["stop"]:
                    #         breakpoint()
                    xp.cuda.runtime.deviceSynchronize()
                    if np.all(np.isnan(inner_here)):
                        breakpoint()
                    
                    try:
                        inner_term[i] = inner_here.get()
                    except AttributeError:
                        inner_term[i] = inner_here

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

                del data_tmp1[gpu_i], data_tmp2[gpu_i], psd_tmp1[gpu_i], psd_tmp2[gpu_i] 

                xp.get_default_memory_pool().free_all_blocks()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

        # if reshape:
        #     inner_term = inner_term.reshape(self.ntemps, self.nwalkers)
            
        # et = time.perf_counter()
        # print("inner prod", et - st)  
        return inner_term
                    
    def get_ll(self, *args, include_psd_info=False, overall_inds=None, **kwargs):
        inner_product = self.get_inner_product(*args, overall_inds=overall_inds, **kwargs)
        ll_out = -1/2 * inner_product

        if include_psd_info:
            ll_out += -self.get_psd_term(overall_inds=overall_inds)
        return ll_out

    def multiply_data(self, val):
        return_to_main = xp.cuda.runtime.getDevice()
        if not isinstance(val, int) and not isinstance(val, float):
            raise NotImplementedError("val must be an int or float.")

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu): 
                for chan in range(len(self.data_list)):
                    self.data_list[chan][gpu_i] *= val

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

    def restore_base_injections(self):
        return_to_main = xp.cuda.runtime.getDevice()
        if self.base_injections is None or self.base_psd is None:
            raise ValueError("Must give base_injections and base_psd kwarg to __init__ to restore.")

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu): 
                for chan in range(len(self.data_list)):
                    tmp = self.data_list[chan][gpu_i].reshape(-1, self.data_length)
                    tmp[:] = xp.asarray(self.base_injections[chan])[None, :]
                    self.data_list[chan][gpu_i] = tmp.flatten()

                    tmp = self.psd_list[chan][gpu_i].reshape(-1, self.data_length)
                    tmp[:] = xp.asarray(self.base_psd[chan])[None, :]
                    self.psd_list[chan][gpu_i] = tmp.flatten()

        for gpu_i, (gpu, gpu_split) in enumerate(zip(self.gpus, self.gpu_splits)):
            with xp.cuda.device.Device(gpu):
                xp.cuda.runtime.deviceSynchronize()

        xp.cuda.runtime.setDevice(return_to_main)
        xp.cuda.runtime.deviceSynchronize()

    def get_injection_inner_product(self, *args, **kwargs):

        inner_out = self.df * 4 * np.sum(
            self.base_injections[0].conj() * self.base_injections[0] / self.base_psd[0]
            + self.base_injections[1].conj() * self.base_injections[1] / self.base_psd[1],
        )
        return inner_out




if __name__ == "__main__":
    ntemps = 2
    nwalkers = 100
    data_length = int(1.6e5)
    nchannels = 2
    df = 3e-8

    data_A = np.ones((ntemps, nwalkers, data_length), dtype=complex)
    data_E = np.ones((ntemps, nwalkers, data_length), dtype=complex)

    psd_A = np.ones((ntemps, nwalkers, data_length), dtype=complex)
    psd_E = np.ones((ntemps, nwalkers, data_length), dtype=complex)

    gpus = [5, 6]
    mg = MultiGPUDataHolder(gpus, data_A, data_E, psd_A, psd_E, df)

    check1 = mg.get_mapped_indices(np.arange(len(mg.overall_indices_flat)))

    mg.map = np.random.choice(mg.overall_indices_flat, len(mg.overall_indices_flat), replace=False) 

    check2 = mg.get_mapped_indices(np.arange(len(mg.overall_indices_flat)))

    check3 = mg.get_ll()
    breakpoint()
    

    


import warnings
from eryn.state import Branch, BranchSupplimental

import numpy as np

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError):
    pass


class Likelihood(object):
    def __init__(
        self,
        template_model,
        num_channels,
        dt=None,
        df=None,
        f_arr=None,
        parameter_transforms=None,
        use_gpu=False,
        vectorized=False,
        separate_d_h=False,
        return_cupy=False,
        fill_data_noise=False,
        transpose_params=False,
        subset=None,
        adjust_psd=False,
    ):
        self.subset = subset
        self.adjust_psd = adjust_psd
        self.transpose_params = transpose_params
        self.template_model = template_model

        self.parameter_transforms = parameter_transforms

        self.fill_data_noise = fill_data_noise

        self.use_gpu = use_gpu
        self.vectorized = vectorized

        self.num_channels = num_channels

        self.separate_d_h = separate_d_h

        if dt is None and df is None and f_arr is None:
            raise ValueError("Must provide dt, df or f_arr.")

        self.dt, self.df, self.f_arr = dt, df, f_arr

        if df is not None or f_arr is not None:
            self.frequency_domain = True
        else:
            self.frequency_domain = False

        self.return_cupy = return_cupy

        self.noise_has_been_added = False
        self._specific_likelihood_setup()

    def _specific_likelihood_setup(self):
        if isinstance(self.template_model, list):
            raise ValueError("For single likelihood, template model cannot be a list.")
        if hasattr(self.template_model, "get_ll"):
            self.get_ll = self.template_model.get_ll
            self.like_here = False

        else:
            self.fill_data_noise = False
            self.like_here = True

    # TODO: add previously injected signal from a file for example
    # TODO: add SNR scaling, need to read out new distance
    def inject_signal(
        self,
        data_stream=None,
        params=None,
        waveform_kwargs={},
        noise_fn=None,
        noise_args=[],
        noise_kwargs={},
        add_noise=False,
    ):
        xp = cp if self.use_gpu else np

        if params is not None:
            if self.parameter_transforms is not None:
                key = list(self.parameter_transforms.keys())[0]
                params = self.parameter_transforms[key].both_transforms(params)

            injection_channels = xp.asarray(
                self.template_model(*params, **waveform_kwargs)
            )
            try:
                injection_channels = injection_channels.get()

            except AttributeError:
                pass

        elif data_stream is not None:
            if isinstance(data_stream, list) is False:
                raise ValueError("If data_stream is provided, it must be as a list.")
            try:
                injection_channels = xp.asarray(data_stream).get()
            except AttributeError:
                injection_channels = np.asarray(data_stream)

        else:
            raise ValueError(
                "Must provide data_stream or params kwargs to inject signal."
            )

        self.injection_length = len(injection_channels[0])

        for inj in injection_channels:
            if len(inj) != self.injection_length:
                raise ValueError("Length of all injection channels must match.")

        if len(injection_channels) != self.num_channels:
            raise ValueError(
                "Number of channels from template_model does not match number of channels declare by user."
            )

        if isinstance(noise_fn, list):
            if len(noise_fn) != 1 and len(noise_fn) != self.num_channels:
                raise ValueError(
                    "Number of noise functions does not match number of channels declared by user."
                )

            elif len(noise_fn) == 1:
                noise_fn = [noise_fn[0] for _ in range(self.num_channels)]

        else:
            noise_fn = [noise_fn for _ in range(self.num_channels)]

        if isinstance(noise_kwargs, list):
            if len(noise_kwargs) != 1 and len(noise_kwargs) != self.num_channels:
                raise ValueError(
                    "Number of noise kwargs does not match number of channels declared by user."
                )
            elif len(noise_kwargs) == 1:
                noise_kwargs = [noise_kwargs[0] for _ in range(self.num_channels)]

        else:
            noise_kwargs = [noise_kwargs for _ in range(self.num_channels)]

        if isinstance(noise_args, list):
            if len(noise_args) != 1 and len(noise_args) != self.num_channels:
                raise ValueError(
                    "Number of noise args does not match number of channels declared by user."
                )
            elif len(noise_args) == 1:
                noise_args = [noise_args[0] for _ in range(self.num_channels)]

        else:
            noise_args = [noise_args for _ in range(self.num_channels)]

        if self.frequency_domain:
            if self.df is not None and self.f_arr is None:
                df = self.df
                can_add_noise = True
                freqs = np.arange(self.injection_length) * df
                injection_channels = [inj for inj in injection_channels]

            else:
                freqs = self.f_arr
                can_add_noise = False
                self.df = (freqs[1] - freqs[0]).item()

        else:
            dt = self.dt
            freqs = np.fft.rfftfreq(self.injection_length, dt)
            can_add_noise = True
            self.df = df = 1.0 / (self.injection_length * dt)

        if self.frequency_domain is False:
            injection_channels = [np.fft.rfft(inj) * dt for inj in injection_channels]

        if not self.adjust_psd:
            psd = [
                noise_fn_temp(freqs, *noise_args_temp, **noise_kwargs_temp)
                for noise_fn_temp, noise_args_temp, noise_kwargs_temp in zip(
                    noise_fn, noise_args, noise_kwargs
                )
            ]

            if np.isnan(psd[0][0]):
                for i in range(len(psd)):
                    psd[i][0] = 1e100

            diff_freqs = np.zeros_like(freqs)
            diff_freqs[1:] = np.diff(freqs)
            diff_freqs[0] = diff_freqs[1]

            self.base_injections = injection_channels
            if add_noise and can_add_noise and self.noise_has_been_added is False:
                raise NotImplementedError
                norm = 0.5 * (1.0 / df) ** 0.5
                noise_to_add = [
                    psd_temp ** (1 / 2)
                    * (
                        np.random.normal(0, norm, len(freqs))
                        + 1j * np.random.normal(0, norm, len(freqs))
                    )
                    for psd_temp in psd
                ]

                self.noise_to_add = noise_to_add

                injection_channels = [
                    inj + noise for inj, noise in zip(injection_channels, noise_to_add)
                ]

                # TODO: need to check this
                self.noise_likelihood_factor = np.sum(
                    [
                        1.0 / 2.0 * (2 * np.pi) * np.sum(diff_freqs * psd_temp)
                        for psd_temp in psd
                    ]
                )
                self.noise_has_been_added = True

                self.noise_added_base_injections = injection_channels

            # noise weighting
            # injection_channels = [
            #    inj * (diff_freqs / psd_temp) ** (1 / 2)
            #    for inj, psd_temp in zip(injection_channels, psd)
            # ]

            # self.psd = xp.asarray(
            #    [(diff_freqs / psd_temp) ** (1 / 2) for psd_temp in psd]
            # )
            self.psd = xp.asarray(psd)

            # if self.like_here is False:
            #    self.psd = [xp.asarray(nf.copy()) for nf in self.psd]

        # if we need to evaluate the psd each time
        else:
            self.noise_fn, self.noise_args, self.noise_kwargs = (
                noise_fn,
                noise_args,
                noise_kwargs,
            )

        self.freqs = xp.asarray(freqs)

        if hasattr(self, "injection_channels") is False:
            self.injection_channels = xp.asarray(injection_channels)
        else:
            self.injection_channels += xp.asarray(injection_channels)

        if self.like_here is False:
            self.injection_channels = [inj.copy() for inj in self.injection_channels]

        self.data_length = len(self.injection_channels[0])

        self.start_freq_ind = int(self.freqs[0] / self.df)

    def get_ll(self, params, data, psd, *args, **kwargs):
        xp = cp if self.use_gpu else np

        if psd is None:
            psd = self.psd

        if data is None:
            data = self.injection_channels

        # TODO: make sure parameter transformations appear in posterior if possible
        num_likes = params.shape[0]
        if self.vectorized:
            template_channels = xp.asarray(
                self.template_model(*params, *args, **kwargs)
            )

        else:
            if isinstance(params[0], np.float64):
                params = [params]
            template_channels = [None for _ in range(len(params))]
            for i, params_i in enumerate(params):
                # np.save("params_last", params_i)
                template_channels[i] = self.template_model(*params_i, *args, **kwargs)

            template_channels = xp.asarray(template_channels)

            # template_channels = xp.asarray(
            #    [self.template_model(*params_i, *args, **kwargs) for params_i in params]
            # )

        if self.frequency_domain is False:
            template_channels = xp.fft.rfft(template_channels, axis=-1) * self.dt

        if psd.ndim == 2:
            psd = psd[xp.newaxis, :, :]

        h = template_channels
        if self.separate_d_h:
            raise NotImplementedError

        else:
            if data.ndim == 2:
                data = data[xp.newaxis, :, :]

            psd[xp.isnan(psd) | xp.isinf(psd)] = 1e20
            # combines all channels into 1D array per likelihood

            d_minus_h = data - h

            # TODO: add inds_slice to here from global
            # start_ind = 1 if np.isnan(psd[0, 0, 0]) else 0

            ll = -(
                1.0
                / 2.0
                * (
                    4.0
                    * self.df
                    * xp.sum(((d_minus_h.conj() * d_minus_h) / psd).real, axis=(1, 2))
                )
            )

            if self.adjust_psd:
                ll += xp.sum(xp.log(psd), axis=(1, 2))

            if self.noise_has_been_added:
                raise NotImplementedError
                # TODO
                ll += self.noise_likelihood_factor

            out = xp.atleast_1d(ll.squeeze())

        if self.use_gpu:
            if self.return_cupy:
                return out
            else:
                try:
                    return out.get()
                except AttributeError:
                    return out

        else:
            return out

    def evaluate_psd(
        self,
        noise_params,
        f_arr=None,
        noise_fn: list = None,
        noise_kwargs: list = None,
        noise_groups=None,
    ):
        xp = cp if self.use_gpu else np

        if noise_groups is None:
            if len(np.unique(noise_groups)) != len(noise_groups):
                raise ValueError(
                    "If providing noise_groups with adjustable leaf count, need to write custom evaluate_psd function."
                )
        if f_arr is None:
            f_arr = self.freqs

        assert isinstance(f_arr, xp.ndarray)

        if noise_fn is None:
            # must be a list
            noise_fn = self.noise_fn

        if noise_kwargs is None:
            # must be a list
            noise_kwargs = self.noise_kwargs

        psd = xp.asarray(
            [
                noise_fn_temp(f_arr, *noise_params, **noise_kwargs_temp)
                for noise_fn_temp, noise_kwargs_temp in zip(noise_fn, noise_kwargs)
            ]
        ).transpose((1, 0, 2))
        return psd

    def __call__(self, params, data=None, psd=None, *args, **kwargs):
        xp = cp if self.use_gpu else np
        if isinstance(params, list):
            if len(params) != 2:
                ValueError(
                    "If providing params for a single source Likelihood, must be an array if just parameters or a list of length 2 where the first entry in the parameter array and the second entry is the parameterization of the noise curve."
                )

            if not self.adjust_psd:
                raise ValueError(
                    "If providing a list with noise parameters, adjust_psd kwarg in __init__ method must be true."
                )
            # must be transpose for noise
            noise_params = params[1].T

            params = params[0]

            if psd is not None:
                raise ValueError(
                    "If providing noise parameters to likelihood, cannot also provide psd kwarg."
                )

        else:
            noise_params = None

        assert isinstance(params, np.ndarray)

        if self.parameter_transforms is not None:
            keys = list(self.parameter_transforms.keys())
            if len(keys) > 1:
                if len(keys) > 2:
                    raise ValueError(
                        "parameter_transforms should only contain transforms for the parameters and the noise parameters."
                    )
                if "noise_params" not in keys or "noise_params" != keys[0]:
                    raise ValueError(
                        "'noise_params' must be the model name given for noise information to maintain consistency. It must be provided in the second position in the parameter_transforms dictionary."
                    )

            params = self.parameter_transforms[keys[0]].both_transforms(params)

            if "noise_params" in keys:
                noise_params = self.parameter_transforms[
                    "noise_params"
                ].both_transforms(noise_params)

        # only has to do with params, not noise params
        if self.transpose_params:
            params = params.T
            subset_axis = 1
        else:
            subset_axis = 0

        num_likes = params.shape[subset_axis]

        inds_likes = np.arange(num_likes)
        if self.subset is not None:
            if not isinstance(self.subset, int):
                raise ValueError("Subset must be int.")
            subset_sep = np.arange(self.subset, num_likes, self.subset)
            inds_subset = np.split(inds_likes, subset_sep)
        else:
            inds_subset = [inds_likes]

        out_ll = []
        for inds in inds_subset:
            if subset_axis == 0:
                args_in = (params[inds],)
            else:
                args_in = (params[:, inds],)

            args_in += args

            if self.fill_data_noise or self.like_here or noise_params is not None:
                if data is None:
                    data = self.injection_channels

                if noise_params is not None:
                    # assumes that there is one set of noise parameters per regular parameters
                    psd = self.evaluate_psd(noise_params[:, inds])
                else:
                    psd = self.psd

                args_in += (data, psd)

            out_ll.append(self.get_ll(*args_in, **kwargs))
        return np.concatenate(out_ll, axis=0)

        # TODO add Subset
        """
        num_inds = len(inds_eval)
            ind_skip = np.arange(self.subset, num_inds, self.subset)
            inds_eval_temp = [inds for inds in np.split(inds_eval, ind_skip)]

            temp = [None for j in range(len(inds_eval_temp))]
            if self.get_d_h:
                d_h = [None for j in range(len(inds_eval_temp))]
                h_h = [None for j in range(len(inds_eval_temp))]

            for j, inds in enumerate(inds_eval_temp):
                if self.add_inds:
                    self.lnlike_kwargs["waveform_kwargs"]["inds"] = inds

                temp[j] = self.lnlike.get_ll(x_in[inds], **self.lnlike_kwargs)
                if self.get_d_h:
                    d_h[j] = self.lnlike.d_h
                    h_h[j] = self.lnlike.h_h
            temp = np.concatenate(temp)
            if self.get_d_h:
                h_h = np.concatenate(h_h)
                d_h = np.concatenate(d_h)

        loglike_vals[inds_eval] = temp
        loglike_vals[np.isnan(loglike_vals)] = np.inf

        if self.get_d_h:
            d_h_vals[inds_eval] = d_h
            h_h_vals[inds_eval] = h_h

        array_1 = (
            -loglike_vals if self.add_prior is False else -loglike_vals + prior_vals
        )
        list_of_arrays = [array_1, prior_vals]
        if self.get_d_h:
            list_of_arrays = list_of_arrays + [d_h_vals, h_h_vals]
        return np.asarray(list_of_arrays).T
    @property
    def d_h(self):
        if self.separate_d_h is False:
            raise ValueError("Cannot get dh term if self.separate_d_h if False.")

        if hasattr(self.template_model, "d_h"):
            return self.template_model.d_h.copy()

        else:
            raise ValueError("Template model does not have the d_h term available.")

    @property
    def h_h(self):
        if self.separate_d_h is False:
            raise ValueError("Cannot get dh term if self.separate_d_h if False.")

        if hasattr(self.template_model, "h_h"):
            return self.template_model.h_h.copy()

        else:
            raise ValueError("Template model does not have the d_h term available.")

        """


class GlobalLikelihood(Likelihood):
    def __init__(
        self,
        *args,
        fill_templates=False,
        **kwargs,
    ):
        super(GlobalLikelihood, self).__init__(*args, **kwargs)
        self.fill_templates = fill_templates

    def _specific_likelihood_setup(self):
        if not isinstance(self.template_model, list):
            self.template_model = [self.template_model]

        if not isinstance(self.parameter_transforms, list):
            self.parameter_transforms = [self.parameter_transforms]

        if not isinstance(self.vectorized, list):
            self.vectorized = [self.vectorized for _ in self.template_model]

        assert (
            len(self.template_model)
            == len(self.parameter_transforms)
            == len(self.vectorized)
        )

        self.like_here = True

    def get_ll(
        self,
        params,
        groups,
        data,
        psd,
        data_length=None,
        start_freq_ind=None,
        args_list=None,
        kwargs_list=None,
        supps=None,
        branch_supps=None,
    ):
        # get supps
        if not isinstance(params, list):
            params = [params]
        if not isinstance(groups, list):
            groups = [groups]

        assert len(groups) == len(params)

        if branch_supps is not None:
            if not isinstance(branch_supps, list):
                branch_supps = [branch_supps]

        else:
            branch_supps = [None for _ in params]

        assert len(groups) == len(branch_supps)

        if args_list is None:
            args_list = [[] for _ in params]

        if kwargs_list is None:
            kwargs_list = [{} for _ in params]

        # TODO: make sure parameter transformations appear in posterior if possible
        total_groups = np.max(np.concatenate(groups)) + 1

        if data_length is not None or start_freq_ind is not None:
            if data_length is None:
                data_length = self.data_length
            elif not isinstance(data_length, int):
                raise ValueError("data_length must be int.")

            if start_freq_ind is None:
                start_freq_ind = self.start_freq_ind
            elif not isinstance(start_freq_ind, int):
                raise ValueError("start_freq_ind must be int.")

            if (start_freq_ind - self.start_freq_ind) + data_length > self.data_length:
                raise ValueError("start_freq_ind + data_length > full data length.")

            if kwargs_list is None:
                kwargs_list = [{}]

            for kwargs in kwargs_list:
                if isinstance(kwargs, dict):
                    kwargs["start_freq_ind"] = start_freq_ind

        else:
            start_freq_ind = self.start_freq_ind
            data_length = self.data_length

        if data is None:
            data = self.injection_channels[xp.newaxis, :, inds_slice]

        if psd is None:
            psd = self.psd[xp.newaxis, :, inds_slice]

        if supps is None or "data_minus_template" not in supps:
            template_all = xp.zeros(
                (total_groups, self.num_channels, data_length),
                dtype=xp.complex128,
            )

            for i, (
                params_i,
                groups_i,
                args_i,
                kwargs_i,
                tm_i,
                vec_i,
                branch_supp_i,
            ) in enumerate(
                zip(
                    params,
                    groups,
                    args_list,
                    kwargs_list,
                    self.template_model,
                    self.vectorized,
                    branch_supps,
                )
            ):
                # TODO: make fill templates adjustable per model
                if not self.fill_templates:  # False
                    if vec_i:
                        template_channels = xp.asarray(
                            tm_i(params_i, *args_i, **kwargs_i)
                        )

                    else:
                        template_channels = xp.asarray(
                            [
                                tm_i(params_ij, *args_i, **kwargs_i)
                                for params_ij in params_i.T
                            ]
                        )

                    if self.frequency_domain is False:
                        # TODO: vectorize this
                        # 2: is removal of DC component + right summation approximation
                        template_channels = (
                            xp.fft.rfft(template_channels, axis=-1) * self.dt
                        )

                    # TODO: could put this in c?
                    for group_ij in np.unique(groups_i):
                        inds1 = np.where(groups_i == group_ij)
                        template_all[group_ij] += template_channels[inds1].sum(axis=0)

                else:  # model will fill templates
                    kwargs_i_in = kwargs_i.copy()
                    if branch_supp_i is not None:
                        kwargs_i_in["branch_supps"] = branch_supp_i
                    if supps is not None:
                        kwargs_i_in["supps"] = supps

                    if vec_i:
                        tm_i.generate_global_template(
                            params_i, groups_i, template_all, *args_i, **kwargs_i_in
                        )

                    else:
                        for params_ij, groups_ij in zip(params_i.T, groups_i):
                            tm_i.generate_global_template(
                                params_ij,
                                groups_ij,
                                template_all,
                                *args_i,
                                **kwargs_i_in,
                            )
            breakpoint()
            # accelerate ?
            d_minus_h = (data - template_all).reshape(
                total_groups, len(self.injection_channels), -1
            )

        else:
            d_minus_h = supps["data_minus_template"]

        start_here = start_freq_ind - self.start_freq_ind
        end_here = start_here + data_length
        inds_slice = slice(start_here, end_here)

        # avoid f = 0
        start_ind = 1 if np.isnan(psd[0, 0, inds_slice][0]) else 0

        self.signal_ll = -(
            1.0
            / 2.0
            * (
                4.0
                * self.df
                * xp.sum(
                    (
                        d_minus_h[:, :, start_ind:].conj() * d_minus_h[:, :, start_ind:]
                    ).real
                    / psd[:, :, start_ind:],
                    axis=(1, 2),
                )
            )
        )

        ll = self.signal_ll.copy()
        if self.adjust_psd:
            self.noise_ll = -xp.sum(xp.log(psd), axis=(1, 2))
            ll += self.noise_ll

        out = xp.atleast_1d(ll.squeeze())
        if xp.any(xp.isnan(out)):
            breakpoint()
        if self.use_gpu:
            if self.return_cupy:
                return out
            else:
                try:
                    return out.get()
                except AttributeError:
                    return out

        else:
            return out

    def __call__(self, params, groups, *args, data=None, psd=None, **kwargs):
        if isinstance(params, np.ndarray):
            params = [params]
        elif not isinstance(params, list):
            raise ValueError("params must be np.ndarray or list of np.ndarray.")

        if isinstance(groups, np.ndarray):
            groups = [groups]
        elif not isinstance(groups, list):
            raise ValueError("groups must be np.ndarray or list of np.ndarray.")

        # if np.any(np.abs(params[0][:, 4]) > 1.0) or np.any(np.abs(params[0][:, 7]) > 1.0):
        #    breakpoint()

        if self.parameter_transforms is not None:
            for i, (params_i, transform_i) in enumerate(
                zip(params, self.parameter_transforms)
            ):
                params[i] = transform_i.both_transforms(params_i.copy())

        else:
            params = [params_i.T for params_i in params]

        for par_i in params:
            if np.any(np.isnan(par_i)):
                breakpoint()

        if self.adjust_psd:
            assert len(params) > 1
            assert len(groups) > 1

            noise_params = params[-1]
            noise_groups = groups[-1]

            if "branch_supps" in kwargs:
                assert len(kwargs["branch_supps"]) == len(groups)
                noise_supps = kwargs["branch_supps"][-1]
                kwargs["branch_supps"] = kwargs["branch_supps"][:-1]

            params = params[:-1]
            groups = groups[:-1]

        else:
            noise_params = None

        args_in = [params] + [groups] + list(args)

        if noise_params is not None:
            psd = self.evaluate_psd(noise_params.T, noise_groups=noise_groups)
        else:
            if psd is None:
                psd = self.psd
            if self.like_here:
                if isinstance(psd, list):
                    psd = xp.asarray(psd)[None, :, :]  # .transpose(1, 0, 2)

        args_in += args

        if self.fill_data_noise or self.like_here or noise_params is not None:
            if data is None:
                data = self.injection_channels

            args_in += [data, psd]

        return self.get_ll(*args_in, **kwargs)

        # TODO add Subset
        """
        num_inds = len(inds_eval)
            ind_skip = np.arange(self.subset, num_inds, self.subset)
            inds_eval_temp = [inds for inds in np.split(inds_eval, ind_skip)]

            temp = [None for j in range(len(inds_eval_temp))]
            if self.get_d_h:
                d_h = [None for j in range(len(inds_eval_temp))]
                h_h = [None for j in range(len(inds_eval_temp))]

            for j, inds in enumerate(inds_eval_temp):
                if self.add_inds:
                    self.lnlike_kwargs["waveform_kwargs"]["inds"] = inds

                temp[j] = self.lnlike.get_ll(x_in[inds], **self.lnlike_kwargs)
                if self.get_d_h:
                    d_h[j] = self.lnlike.d_h
                    h_h[j] = self.lnlike.h_h
            temp = np.concatenate(temp)
            if self.get_d_h:
                h_h = np.concatenate(h_h)
                d_h = np.concatenate(d_h)

        loglike_vals[inds_eval] = temp
        loglike_vals[np.isnan(loglike_vals)] = np.inf

        if self.get_d_h:
            d_h_vals[inds_eval] = d_h
            h_h_vals[inds_eval] = h_h

        array_1 = (
            -loglike_vals if self.add_prior is False else -loglike_vals + prior_vals
        )
        list_of_arrays = [array_1, prior_vals]
        if self.get_d_h:
            list_of_arrays = list_of_arrays + [d_h_vals, h_h_vals]
        return np.asarray(list_of_arrays).T
    @property
    def d_h(self):
        if self.separate_d_h is False:
            raise ValueError("Cannot get dh term if self.separate_d_h if False.")

        if hasattr(self.template_model, "d_h"):
            return self.template_model.d_h.copy()

        else:
            raise ValueError("Template model does not have the d_h term available.")

    @property
    def h_h(self):
        if self.separate_d_h is False:
            raise ValueError("Cannot get dh term if self.separate_d_h if False.")

        if hasattr(self.template_model, "h_h"):
            return self.template_model.h_h.copy()

        else:
            raise ValueError("Template model does not have the d_h term available.")

        """



import warnings
from abc import ABC
from typing import Any, Tuple, Optional, List, Dict

import math
import numpy as np
from numpy.typing import ArrayLike
from scipy import interpolate

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError):
    import numpy as cp

from . import detector as lisa_models
from .utils.utility import AET
from .utils.constants import *


class StochasticContribution(ABC):
    """Base Class for Stochastic Contributions to the PSD."""

    ndim = None
    added_stochastic_list = []

    @classmethod
    def _check_ndim(cls, params: ArrayLike) -> None:
        """Check the dimensionality of the parameters matches the model.

        Args:
            params: Parameters for stochastic model.

        """
        if cls.ndim is None:
            raise ValueError(
                "When subclassing the StochasticContribution class, must set `ndim` as a static attribute."
            )

        if len(params) != cls.ndim:
            raise ValueError("length of parameters is not equivalent to class ndim.")

    @classmethod
    def get_Sh(
        cls, f: float | np.ndarray, *params: ArrayLike, **kwargs: Any
    ) -> float | np.ndarray:
        """Calculate the power spectral density of the stochastic contribution.

        Args:
            f: Frequency array.
            *params: Parameters for the stochastic model.
            **kwargs: Keyword arguments for the stochastic model.

        """
        if len(cls.added_stochastic_list) > 0:
            cls._check_ndim(params[0])
        return cls.specific_Sh_function(f, *params, **kwargs)

    @staticmethod
    def specific_Sh_function(
        f: float | np.ndarray, *args: Any, **kwargs: Any
    ) -> float | np.ndarray:
        """Calculate the power spectral density contained in a stochastic signal contribution.

        Args:
            f: Frequency array.
            *args: Any arguments for the function.
            **kwargs: Any keyword arguments for the function.

        Returns:
            Power spectral density contained in stochastic signal.

        """
        raise NotImplementedError


class StochasticContributionContainer:
    """Container for multiple Stochastic Contributions

    Args:
        stochastic_contribution_dict: Dictionary with multiple Stochastic entries.
            Keys are the names and values are of type :class:`StochasticContribution`.

    """

    def __init__(
        self, stochastic_contribution_dict: dict[StochasticContribution]
    ) -> None:
        self.stochastic_contribution_dict = stochastic_contribution_dict

    @property
    def stochastic_contribution_dict(self) -> dict[StochasticContribution]:
        """Stochastic contribution storage."""
        return self._stochastic_contribution_dict

    @stochastic_contribution_dict.setter
    def stochastic_contribution_dict(
        self, stochastic_contribution_dict: dict[StochasticContribution]
    ) -> None:
        """Set stochastic_contribution_dict."""
        assert isinstance(stochastic_contribution_dict, dict)
        for key, value in stochastic_contribution_dict.items():
            if not isinstance(value, StochasticContribution):
                raise ValueError(
                    f"Stochastic model {key} is not of type StochasticContribution."
                )
        self._stochastic_contribution_dict = stochastic_contribution_dict

    def get_Sh(
        self, f: float | np.ndarray, params_dict: dict[tuple], kwargs_dict: dict[dict]
    ) -> np.ndarray:
        """Calculate Sh for stochastic contribution.

        Args:
            f: Frequency array.
            params_dict: Dictionary with keys equivalent to ``self.stochastic_contribution_dict.keys()``.
                Values are the parameters for each associated model.
            kwargs_dict: Dictionary with keys equivalent to ``self.stochastic_contribution_dict.keys()``.
                Values are the keyword argument dicts for each associated model.

        Returns:
            Stochastic contribution.

        """
        Sh_out = np.zeros_like(f)
        for key in params_dict:
            stochastic_contrib = self.stochastic_contribution_dict[key]
            Sh_out += stochastic_contrib.get_Sh(
                f, params_dict[key], **(kwargs_dict.get(key, {}))
            )
        return Sh_out

    def __setitem__(self, key: str | int | tuple, val: StochasticContribution) -> None:
        """Set an item by directly indexing the class object."""
        self.stochastic_contribution_dict[key] = val

    def __getitem__(self, key: str | int | tuple) -> StochasticContribution:
        """Get item directly from dictionary."""
        return self.stochastic_contribution_dict[key]


class HyperbolicTangentGalacticForeground(StochasticContribution):
    """Hyperbolic Tangent-based foreground fitting function."""

    ndim = 5

    @staticmethod
    def specific_Sh_function(
        f: float | np.ndarray, amp: float, fk: float, alpha: float, s1: float, s2: float
    ) -> float | np.ndarray:
        """Hyperbolic tangent model 1 for the Galaxy foreground noise

        This model for the PSD contribution from the Galactic foreground noise is given by

        .. math::

            S_\\text{gal} = \\frac{A_\\text{gal}}{2}e^{-s_1f^\\alpha}f^{-7/3}\\left[ 1 + \\tanh{\\left(-s_2 (f - f_k)\\right)} \\right],

        where :math:`A_\\text{gal}` is the amplitude of the stochastic signal, :math:`f_k` is the knee frequency at which a bend occurs,
        math:`\\alpha` is a power law parameter, :math:`s_1` is a slope parameter below the knee,
        and :math:`s_2` is a slope parameter after the knee.:

        Args:
            f: Frequency array.
            amp: Amplitude parameter for the Galaxy.
            fk: Knee frequency in Hz.
            alpha: Power law parameter.
            s1: Slope parameter below knee.
            s2: Slope parameter above knee.

        Returns:
            PSD of the Galaxy foreground noise

        """
        Sgal = (
            amp
            * np.exp(-(f**alpha) * s1)
            * (f ** (-7.0 / 3.0))
            * 0.5
            * (1.0 + np.tanh(-(f - fk) * s2))
        )

        return Sgal


class FittedHyperbolicTangentGalacticForeground(HyperbolicTangentGalacticForeground):
    # TODO: need to verify this is still working
    ndim = 1
    amp = 3.26651613e-44
    alpha = 1.18300266e00
    # Tobs should be in sec.
    day = 86400.0
    month = day * 30.5
    year = 365.25 * 24.0 * 3600.0  # hard coded for initial fits

    Xobs = [
        1.0 * day,
        3.0 * month,
        6.0 * month,
        1.0 * year,
        2.0 * year,
        4.0 * year,
        10.0 * year,
    ]
    knee = [
        1.15120924e-02,
        4.01884128e-03,
        3.47302482e-03,
        2.77606177e-03,
        2.41178384e-03,
        2.09278117e-03,
        1.57362626e-03,
    ]
    Slope1 = [
        9.41315118e02,
        1.36887568e03,
        1.68729474e03,
        1.76327234e03,
        2.32678814e03,
        3.01430978e03,
        3.74970124e03,
    ]

    Slope2 = [
        1.03239773e02,
        1.03351646e03,
        1.62204855e03,
        1.68631844e03,
        2.06821665e03,
        2.95774596e03,
        3.15199454e03,
    ]
    Tmax = 10 * YRSID_SI

    @classmethod
    def specific_Sh_function(
        cls, f: float | np.ndarray, Tobs: float
    ) -> float | np.ndarray:
        """Fitted hyperbolic tangent model 1 for the Galaxy foreground noise.

        This class fits the parameters for :class:`HyperbolicTangentGalacticForeground`
        using analytic estimates from (# TODO). The fit is a function of time, so the user
        inputs ``Tobs``.

        # Sgal_1d = 2.2e-44*np.exp(-(fr**1.2)*0.9e3)*(fr**(-7./3.))*0.5*(1.0 + np.tanh(-(fr-1.4e-2)*0.7e2))
        # Sgal_3m = 2.2e-44*np.exp(-(fr**1.2)*1.7e3)*(fr**(-7./3.))*0.5*(1.0 + np.tanh(-(fr-4.8e-3)*5.4e2))
        # Sgal_1y = 2.2e-44*np.exp(-(fr**1.2)*2.2e3)*(fr**(-7./3.))*0.5*(1.0 + np.tanh(-(fr-3.1e-3)*1.3e3))
        # Sgal_2y = 2.2e-44*np.exp(-(fr**1.2)*2.2e3)*(fr**(-7./3.))*0.5*(1.0 + np.tanh(-(fr-2.3e-3)*1.8e3))
        # Sgal_4y = 2.2e-44*np.exp(-(fr**1.2)*2.9e3)*(fr**(-7./3.))*0.5*(1.0 + np.tanh(-(fr-2.0e-3)*1.9e3))

        Args:
            f: Frequency array.
            Tobs: Observation time in years.

        Returns:
            PSD of the Galaxy foreground noise

        """

        if Tobs > cls.Tmax:
            raise ValueError(
                "Tobs is greater than the maximum allowable fit which is 10 years."
            )

        # Interpolate
        tck1 = interpolate.splrep(cls.Xobs, cls.Slope1, s=0, k=1)
        tck2 = interpolate.splrep(cls.Xobs, cls.knee, s=0, k=1)
        tck3 = interpolate.splrep(cls.Xobs, cls.Slope2, s=0, k=1)
        s1 = interpolate.splev(Tobs, tck1, der=0).item()
        fk = interpolate.splev(Tobs, tck2, der=0).item()
        s2 = interpolate.splev(Tobs, tck3, der=0).item()

        return HyperbolicTangentGalacticForeground.specific_Sh_function(
            f, cls.amp, fk, cls.alpha, s1, s2
        )


__stock_gb_stochastic_options__ = [
    "HyperbolicTangentGalacticForeground",
    "FittedHyperbolicTangentGalacticForeground",
]


def get_stock_gb_stochastic_options() -> List[StochasticContribution]:
    """Get stock options for stochastic contributions.

    Returns:
        List of stock stochastic options.

    """
    return __stock_gb_stochastic_options__



# -*- coding: utf-8 -*-

from copy import deepcopy
from inspect import Attribute
import numpy as np
from scipy import stats
import warnings
import time

try:
    import cupy as xp

    gpu_available = True
except ModuleNotFoundError:
    import numpy as xp

    gpu_available = False

from eryn.moves import StretchMove
from eryn.prior import ProbDistContainer
from eryn.state import State


__all__ = ["GBSpecialStretchMove"]

# MHMove needs to be to the left here to overwrite GBBruteRejectionRJ RJ proposal method
class GBForegroundSpecialMove(StretchMove):
    """
    """

    def __init__(
        self,
        gb,
        priors,
        start_freq_ind,
        data_length,
        mgh,
        fd,
        band_edges,
        *args,
        waveform_kwargs={},
        noise_kwargs={},
        parameter_transforms=None,
        search=False,
        search_samples=None,
        search_snrs=None,
        search_snr_lim=None,
        search_snr_accept_factor=1.0,
        take_max_ll=False,
        global_template_builder=None,
        psd_func=None,
        provide_betas=False,
        alternate_priors=None,
        batch_size=5,
        **kwargs
    ):
        StretchMove.__init__(self, **kwargs)  

        self.time = 0
        self.greater_than_1e0 = 0
        self.name = "GBForegroundSpecialMove".lower()

        # TODO: make priors optional like special generate function? 
        for key in priors:
            if not isinstance(priors[key], ProbDistContainer):
                raise ValueError("Priors need to be eryn.priors.ProbDistContainer object.")
        self.priors = priors
        self.gb = gb
        self.provide_betas = provide_betas
        self.batch_size = batch_size
        self.stop_here = True

        # use gpu from template generator
        self.use_gpu = gb.use_gpu
        if self.use_gpu:
            self.xp = xp
            self.mempool = self.xp.get_default_memory_pool()

        else:
            self.xp = np


        self.band_edges = band_edges
        self.num_bands = len(band_edges) - 1
        self.start_freq_ind = start_freq_ind
        self.data_length = data_length
        self.waveform_kwargs = waveform_kwargs
        self.noise_kwargs = noise_kwargs
        self.parameter_transforms = parameter_transforms
        self.psd_func = psd_func
        self.fd = fd
        self.df = (fd[1] - fd[0]).item()
        self.mgh = mgh
        self.search = search
        self.global_template_builder = global_template_builder

        if search_snrs is not None:
            if search_snr_lim is None:
                search_snr_lim = 0.1

            assert len(search_samples) == len(search_snrs)

        self.search_samples = search_samples
        self.search_snrs = search_snrs
        self.search_snr_lim = search_snr_lim
        self.search_snr_accept_factor = search_snr_accept_factor

        self.take_max_ll = take_max_ll     

    def propose(self, model, state):
        """Use the move to generate a proposal and compute the acceptance

        Args:
            model (:class:`eryn.model.Model`): Carrier of sampler information.
            state (:class:`State`): Current state of the sampler.

        Returns:
            :class:`State`: State of sampler after proposal is complete.

        """
        # st = time.perf_counter()
        self.xp.cuda.runtime.setDevice(self.mgh.gpus[0])

        # np.random.seed(10)
        #print("start stretch")
        #st = time.perf_counter()
        # Check that the dimensions are compatible.
        ntemps, nwalkers, nleaves, ndim = state.branches["galfor"].shape

        self.nwalkers = nwalkers
        # TODO: deal with more intensive acceptance fractions
        # Run any move-specific setup.
        self.setup(state.branches)

        new_state = State(state)  # , copy=True)
        self.mempool.free_all_blocks()
        
        self.mgh.map = new_state.supplimental.holder["overall_inds"].flatten()

        # data should not be whitened
        
        # Split the ensemble in half and iterate over these two halves.
        accepted = np.zeros((ntemps, nwalkers), dtype=bool)

        ntemps, nwalkers, nleaves_max, ndim = state.branches_coords["gb"].shape

        split_inds = np.zeros(nwalkers, dtype=int)
        split_inds[1::2] = 1
        np.random.shuffle(split_inds)

        current_coords_galfor = new_state.branches["galfor"].coords.copy()
        """et = time.perf_counter()
        print("setup", (et - st))
        st = time.perf_counter()"""
        for split in range(2):
            # st = time.perf_counter()
            split_here = split_inds == split
            walkers_keep = self.xp.arange(nwalkers)[split_here]
                
            points_to_move = {key: new_state.branches[key].coords[:, split_here] for key in ["psd", "galfor"]}
            points_for_move = {key: [new_state.branches[key].coords[:, ~split_here]] for key in ["psd", "galfor"]}
            
            q, factors = self.get_proposal(
                points_to_move,  
                points_for_move, 
                model.random
            )

            temp_part_general = np.repeat(np.arange(ntemps)[:, None], nwalkers, axis=-1)[:, split_here].flatten()
            walker_part_general = np.tile(np.arange(nwalkers), (ntemps, 1))[:, split_here].flatten()

            # get logp
            logp_here = (self.priors["psd"].logpdf(q["psd"].reshape(-1, q["psd"].shape[-1])) + self.priors["galfor"].logpdf(q["galfor"].reshape(-1, q["galfor"].shape[-1]))).reshape((ntemps, int(nwalkers / 2)))

            prev_logp_here = (self.priors["psd"].logpdf(new_state.branches_coords["psd"][:, split_here].reshape(-1, q["psd"].shape[-1])) + self.priors["galfor"].logpdf(new_state.branches_coords["galfor"][:, split_here].reshape(-1, q["galfor"].shape[-1]))).reshape((ntemps, int(nwalkers / 2)))

            bad = np.isinf(logp_here.flatten())

            data_index_tmp = np.asarray((temp_part_general * self.nwalkers + walker_part_general).astype(xp.int32))

            data_index = self.mgh.get_mapped_indices(data_index_tmp)

            data_index_in = data_index[~bad]

            psd_params = q["psd"].reshape(-1, q["psd"].shape[-1])[~bad]
            foreground_params = q["galfor"].reshape(-1, q["galfor"].shape[-1])[~bad]
    
            self.mgh.set_psd_vals(psd_params, foreground_params=foreground_params, overall_inds=data_index_in)

            logl_temp = self.mgh.get_ll(include_psd_info=True, overall_inds=data_index_in)

            logl = np.full((ntemps, int(nwalkers / 2)), -1e300)
            logl[~bad.reshape(ntemps, -1)] = logl_temp

            prev_logl = new_state.log_like[:, split_here]
            prev_logp = new_state.log_prior[:, split_here]

            logp = prev_logp + logp_here - prev_logp_here

            logP = self.compute_log_posterior(logl, logp)
            prev_logP = self.compute_log_posterior(prev_logl, prev_logp)

            lnpdiff = factors + logP - prev_logP

            keep = lnpdiff > np.log(model.random.rand(ntemps, int(nwalkers / 2)))

            temp_inds_keep = temp_part_general[keep.flatten()]
            walker_inds_keep = walker_part_general[keep.flatten()]

            accepted[temp_inds_keep, walker_inds_keep] = True

            new_state.log_like[temp_inds_keep, walker_inds_keep] = logl[keep]
            new_state.log_prior[temp_inds_keep, walker_inds_keep] = logp[keep]

            for key in ["psd", "galfor"]:
                new_state.branches[key].coords[temp_inds_keep, walker_inds_keep, np.zeros_like(walker_inds_keep)] = q[key][keep][:, 0]

            temp_inds_fix = temp_part_general[~keep.flatten()]
            walker_inds_fix = walker_part_general[~keep.flatten()]

            # return unaccepted psds
            data_index_fix = data_index[~keep.flatten()]
            psd_params_fix = new_state.branches_coords["psd"][(temp_inds_fix, walker_inds_fix)][:, 0]
            foreground_params_fix = new_state.branches_coords["galfor"][(temp_inds_fix, walker_inds_fix)][:, 0]

            self.mgh.set_psd_vals(psd_params_fix, foreground_params=foreground_params_fix, overall_inds=data_index_fix)

        self.accepted += accepted.astype(int)
        self.num_proposals += 1

        self.mempool.free_all_blocks()

        if self.time % 200 == 0:
            ll_after = self.mgh.get_ll(include_psd_info=True).flatten()[new_state.supplimental[:]["overall_inds"]].reshape(ntemps, nwalkers)
            check = np.abs(new_state.log_like - ll_after).max()
            if check  > 1e-3:
                breakpoint()
                self.mgh.restore_base_injections()

                for name in new_state.branches.keys():
                    if name not in ["gb", "gb"]:
                        continue
                    new_state_branch = new_state.branches[name]
                    coords_here = new_state_branch.coords[new_state_branch.inds]
                    ntemps, nwalkers, nleaves_max_here, ndim = new_state_branch.shape
                    try:
                        group_index = self.xp.asarray(
                            self.mgh.get_mapped_indices(
                                np.repeat(np.arange(ntemps * nwalkers).reshape(ntemps, nwalkers, 1), nleaves_max, axis=-1)[new_state_branch.inds]
                            ).astype(self.xp.int32)
                        )
                    except IndexError:
                        breakpoint()
                    coords_here_in = self.parameter_transforms.both_transforms(coords_here, xp=np)

                    waveform_kwargs_fill = self.waveform_kwargs.copy()
                    waveform_kwargs_fill["start_freq_ind"] = self.start_freq_ind

                    if "N" in waveform_kwargs_fill:
                        waveform_kwargs_fill.pop("N")
                        
                    self.mgh.multiply_data(-1.)
                    self.gb.generate_global_template(coords_here_in, group_index, self.mgh.data_list, data_length=self.data_length, data_splits=self.mgh.gpu_splits, batch_size=1000, **waveform_kwargs_fill)
                    self.mgh.multiply_data(-1.)

                ll_after2 = self.mgh.get_ll(use_cpu=True).flatten()[new_state.supplimental[:]["overall_inds"]].reshape(ntemps, nwalkers)
                new_state.log_like = ll_after2
                   
            """
            data_minus_template = self.xp.concatenate(
                [
                    tmp.reshape(ntemps, nwalkers, 1, self.data_length) for tmp in data_minus_template_in_swap
                ],
                axis=2
            )
            del data_minus_template_in_swap

            psd = self.xp.concatenate(
                [
                    tmp.reshape(ntemps * nwalkers, 1, self.data_length) for tmp in psd_in_swap
                ],
                axis=1
            )
            del psd_in_swap
            self.mempool.free_all_blocks()

            new_state.supplimental.holder["data_minus_template"] = data_minus_template

            lp_after = model.compute_log_prior_fn(new_state.branches_coords, inds=new_state.branches_inds)
            
            ll_after = (-1/2 * 4 * self.df * self.xp.sum(data_minus_template.conj() * data_minus_template / self.xp.asarray(self.psd), axis=(2, 3))).get()  # model.compute_log_like_fn(new_state.branches_coords, inds=new_state.branches_inds, logp=lp_after, supps=new_state.supplimental, branch_supps=new_state.branches_supplimental)
            #check = -1/2 * 4 * self.df * self.xp.sum(data_minus_template.conj() * data_minus_template / self.xp.asarray(self.psd), axis=(2, 3))
            #check2 = -1/2 * 4 * self.df * self.xp.sum(tmp.conj() * tmp / self.xp.asarray(self.psd), axis=(2, 3))
            #print(np.abs(new_state.log_like - ll_after[0]).max())

            # if any are even remotely getting to be different, reset all (small change)
            if np.abs(new_state.log_like - ll_after).max() > 1e-1:
                if np.abs(new_state.log_like - ll_after).max() > 1e0:
                    self.greater_than_1e0 += 1
                    print("Greater:", self.greater_than_1e0)
                breakpoint()
                fix_here = np.abs(new_state.log_like - ll_after) > 1e-6
                data_minus_template_old = data_minus_template.copy()
                data_minus_template = self.xp.zeros_like(data_minus_template_old)
                data_minus_template[:] = self.xp.asarray(self.data)[None, None]
                templates = self.xp.zeros_like(data_minus_template).reshape(-1, 2, data_minus_template.shape[-1])
                for name in new_state.branches.keys():
                    if name not in ["gb", "gb"]:
                        continue
                    new_state_branch = new_state.branches[name]
                    coords_here = new_state_branch.coords[new_state_branch.inds]
                    ntemps, nwalkers, nleaves_max_here, ndim = new_state_branch.shape
                    try:
                        group_index = np.repeat(np.arange(ntemps * nwalkers).reshape(ntemps, nwalkers, 1), nleaves_max, axis=-1)[new_state_branch.inds]
                    except IndexError:
                        breakpoint()
                    coords_here_in = self.parameter_transforms.both_transforms(coords_here, xp=np)

                    self.gb.generate_global_template(coords_here_in, group_index, templates, batch_size=1000, **self.waveform_kwargs)

                data_minus_template -= templates.reshape(ntemps, nwalkers, 2, templates.shape[-1])

                new_like = -1 / 2 * 4 * self.df * self.xp.sum(data_minus_template.conj() * data_minus_template / psd, axis=(2, 3)).real.get()
            
                new_like += self.noise_ll
                new_state.log_like[:] = new_like.reshape(ntemps, nwalkers)

            self.mempool.free_all_blocks()
            data_minus_template_in_swap = [data_minus_template[:,:, 0, :].flatten().copy(), data_minus_template[:,:, 1, :].flatten().copy()]
            del data_minus_template

            psd_in_swap = [psd[:, 0, :].flatten().copy(), psd[:, 1, :].flatten().copy()]
            self.mempool.free_all_blocks()
            del psd
            self.mempool.free_all_blocks()
            """
        
        self.mempool.free_all_blocks()

        if self.temperature_control is not None:
            new_state = self.temperature_control.temper_comps(new_state, adapt=False)
            
            """# new_state, accepted = self.temperature_control.temper_comps(new_state, accepted)
            self.swaps_accepted = np.zeros(ntemps - 1)
            self.attempted_swaps = np.zeros(ntemps - 1)
            betas = self.temperature_control.betas
            for i in range(ntemps - 1, 0, -1):
                bi = betas[i]
                bi1 = betas[i - 1]

                dbeta = bi1 - bi

                iperm = np.random.permutation(nwalkers)
                i1perm = np.random.permutation(nwalkers)

                # need to calculate switch likelihoods

                coords_iperm = new_state.branches["gb"].coords[i, iperm]
                coords_i1perm = new_state.branches["gb"].coords[i - 1, i1perm]

                N_vals_iperm = new_state.branches["gb"].branch_supplimental.holder["N_vals"][i, iperm]

                N_vals_i1perm = new_state.branches["gb"].branch_supplimental.holder["N_vals"][i - 1, i1perm]

                f_test_i = coords_iperm[None, :, :, 1] / 1e3
                f_test_2_i = coords_i1perm[None, :, :, 1] / 1e3
                
                fix_f_test_i = (np.abs(f_test_i - f_test_2_i) > (self.df * N_vals_iperm * 1.5))

                if hasattr(self, "keep_bands") and self.keep_bands is not None:
                    band_indices = np.searchsorted(self.band_edges, f_test_i.flatten()).reshape(f_test_i.shape) - 1
                    keep_bands = self.keep_bands
                    assert isinstance(keep_bands, np.ndarray)
                    fix_f_test_i[~np.in1d(band_indices, keep_bands).reshape(band_indices.shape)] = True


                groups = get_groups_from_band_structure(f_test_i, self.band_edges, f0_2=f_test_2_i, xp=np, num_groups_base=3, fix_f_test=fix_f_test_i)

                unique_groups, group_len = np.unique(groups.flatten(), return_counts=True)

                # remove information about the bad "-1" group
                for check_val in [-1, -2]:
                    group_len = np.delete(group_len, unique_groups == check_val)
                    unique_groups = np.delete(unique_groups, unique_groups == check_val)

                # needs to be max because some values may be missing due to evens and odds
                num_groups = unique_groups.max().item() + 1

                for group_iter in range(num_groups):
                    # st = time.perf_counter()
                    # sometimes you will have an extra odd or even group only
                    # the group_iter may not match the actual running group number in this case
                    if group_iter not in groups:
                        continue
                        
                    group = [grp[i:i+1][groups == group_iter].flatten() for grp in group_temp_finder]

                    # st = time.perf_counter()
                    temp_inds_back, walkers_inds_back, leaf_inds = [self.xp.asarray(grp) for grp in group] 

                    temp_inds_i = temp_inds_back.copy()
                    walkers_inds_i = walkers_inds_back.copy()

                    temp_inds_i1 = temp_inds_back.copy()
                    walkers_inds_i1 = walkers_inds_back.copy()

                    temp_inds_i[:] = i
                    walkers_inds_i[:] = self.xp.asarray(iperm)[walkers_inds_back]

                    temp_inds_i1[:] = i - 1
                    walkers_inds_i1[:] = self.xp.asarray(i1perm)[walkers_inds_back]

                    group_here_i = (temp_inds_i, walkers_inds_i, leaf_inds)

                    group_here_i1 = (temp_inds_i1, walkers_inds_i1, leaf_inds)

                    # factors_here = factors[group_here]
                    old_points = self.xp.asarray(new_state.branches["gb"].coords)[group_here_i]
                    new_points = self.xp.asarray(new_state.branches["gb"].coords)[group_here_i1]

                    N_vals_here_i = N_vals[group_here_i]
                    
                    log_like_tmp = self.xp.asarray(new_state.log_like.copy())
                    log_prior_tmp = self.xp.asarray(new_state.log_prior.copy())

                    delta_logl_i = self.run_swap_ll(None, old_points, new_points, group_here_i, N_vals_here_i, waveform_kwargs_now, None, log_like_tmp, log_prior_tmp, return_at_logl=True)

                    # factors_here = factors[group_here]
                    old_points[:] = self.xp.asarray(new_state.branches["gb"].coords)[group_here_i1]
                    new_points[:] = self.xp.asarray(new_state.branches["gb"].coords)[group_here_i]

                    N_vals_here_i1 = N_vals[group_here_i1]
                    
                    log_like_tmp[:] = self.xp.asarray(new_state.log_like.copy())
                    log_prior_tmp[:] = self.xp.asarray(new_state.log_prior.copy())

                    delta_logl_i1 = self.run_swap_ll(None, old_points, new_points, group_here_i1, N_vals_here_i1, waveform_kwargs_now, None, log_like_tmp, log_prior_tmp, return_at_logl=True)

                    paccept = dbeta * 1. / 2. * (delta_logl_i - delta_logl_i1)
                    raccept = np.log(np.random.uniform(size=paccept.shape[0]))

                    # How many swaps were accepted?
                    sel = paccept > self.xp.asarray(raccept)

                    inds_i_swap = tuple([tmp[sel].get() for tmp in list(group_here_i)])
                    inds_i1_swap = tuple([tmp[sel].get() for tmp in list(group_here_i1)])
                    
                    group_index_i = self.xp.asarray(
                        self.mgh.get_mapped_indices(
                            temp_inds_i[sel] + nwalkers * walkers_inds_i[sel]
                        )
                    ).astype(self.xp.int32)

                    group_index_i1 = self.xp.asarray(
                        self.mgh.get_mapped_indices(
                            temp_inds_i1[sel] + nwalkers * walkers_inds_i1[sel]
                        )
                    ).astype(self.xp.int32)

                    N_vals_i = N_vals[inds_i_swap]
                    params_i = self.xp.asarray(new_state.branches["gb"].coords)[inds_i_swap]
                    params_i1 = self.xp.asarray(new_state.branches["gb"].coords)[inds_i1_swap]

                    params_generate = self.xp.concatenate([
                        params_i,
                        params_i1,
                        params_i1,  # reverse of above
                        params_i,
                    ], axis=0)

                    params_generate_in = self.parameter_transforms.both_transforms(params_generate, xp=self.xp)

                    group_index_gen = self.xp.concatenate(
                        [
                            group_index_i,
                            group_index_i,
                            group_index_i1,
                            group_index_i1
                        ], dtype=self.xp.int32
                    )

                    factors_multiply_generate = self.xp.concatenate([
                        +1 * self.xp.ones_like(group_index_i, dtype=float),
                        -1 * self.xp.ones_like(group_index_i, dtype=float),
                        +1 * self.xp.ones_like(group_index_i, dtype=float),
                        -1 * self.xp.ones_like(group_index_i, dtype=float),
                    ])

                    N_vals_in_gen = self.xp.concatenate([
                        N_vals_i,
                        N_vals_i,
                        N_vals_i,
                        N_vals_i
                    ])
                    
                    waveform_kwargs_fill = waveform_kwargs_now.copy()
                    waveform_kwargs_fill["start_freq_ind"] = self.start_freq_ind

                    self.gb.generate_global_template(
                        params_generate_in,
                        group_index_gen, 
                        self.mgh.data_list, 
                        N=N_vals_in_gen,
                        data_length=self.data_length, 
                        data_splits=self.mgh.gpu_splits, 
                        factors=factors_multiply_generate,
                        **waveform_kwargs_fill
                    )

                    # update likelihoods

                    # set unaccepted differences to zero
                    accepted_delta_ll_i = delta_logl_i * (sel)
                    accepted_delta_ll_i1 = delta_logl_i1 * (sel)

                    logl_change_contribution = np.zeros_like(log_like_tmp.get())
                    try:
                        in_tuple = (accepted_delta_ll_i[sel].get(), accepted_delta_ll_i1[sel].get(), temp_inds_i[sel].get(), temp_inds_i1[sel].get(), walkers_inds_i[sel].get(), walkers_inds_i[sel].get())
                    except AttributeError:
                        in_tuple = (accepted_delta_ll_i[sel], accepted_delta_ll_i1[sel], temp_inds_i[sel], temp_inds_i1[sel], walkers_inds_i[sel], walkers_inds_i[sel])
                    for j, (dlli, dlli1, ti, ti1, wi, wi1) in enumerate(zip(*in_tuple)):
                        logl_change_contribution[ti, wi] += dlli
                        logl_change_contribution[ti1, wi1] += dlli1

                    log_like_tmp[:] += self.xp.asarray(logl_change_contribution)

                    tmp_swap = new_state.branches["gb"].coords[inds_i_swap]
                    new_state.branches["gb"].coords[inds_i_swap] = new_state.branches["gb"].coords[inds_i1_swap]

                    new_state.branches["gb"].coords[inds_i1_swap] = tmp_swap

                    tmp_swap = new_state.branches["gb"].branch_supplimental[inds_i_swap]

                    new_state.branches["gb"].branch_supplimental[inds_i_swap] = new_state.branches["gb"].branch_supplimental[inds_i1_swap]

                    new_state.branches["gb"].branch_supplimental[inds_i1_swap] = tmp_swap

                    # inds are all non-zero
                    self.swaps_accepted[i - 1] += np.sum(sel)
                    self.attempted_swaps[i - 1] += sel.shape[0]

                    ll_after = self.mgh.get_ll(use_cpu=True).flatten()[new_state.supplimental[:]["overall_inds"]].reshape(ntemps, nwalkers)
                    breakpoint()
                    """
        else:
            self.temperature_control.swaps_accepted = np.zeros((ntemps - 1))
        
        
        if np.any(new_state.log_like > 1e10):
            breakpoint()

        self.time += 1
        #self.xp.cuda.runtime.deviceSynchronize()
        #et = time.perf_counter()
        #print("end stretch", (et - st))

        self.mgh.map = new_state.supplimental.holder["overall_inds"].flatten()

        """et = time.perf_counter()
        print("end", (et - st), group_iter, group_len[group_iter])"""
                    
        # breakpoint()
        return new_state, accepted




import time 

import numpy as np

from eryn.utils.stopping import Stopping
from eryn.utils.utility import thermodynamic_integration_log_evidence


class SNRStopping(Stopping):
    def __init__(self, snr_limit=100.0, verbose=False):
        self.snr_limit = snr_limit
        self.verbose = verbose

    def __call__(self, iter, sample, sampler):

        ind = sampler.get_log_like().argmax()

        log_best = sampler.get_log_like().max()
        snr_best = sampler.get_blobs()[:, :, :, 0].flatten()[ind]
        # d_h_best = sampler.get_blobs()[:, :, :, 1].flatten()[ind]
        # h_h_best = sampler.get_blobs()[:, :, :, 2].flatten()[ind]

        if self.verbose:
            print(
                "snr_best",
                snr_best,
                "limit:",
                self.snr_limit,
                "loglike:",
                log_best,
                # d_h_best,
                # h_h_best,
            )

        if snr_best > self.snr_limit:
            return True

        else:
            return False


class NLeavesSearchStopping:
    def __init__(self, convergence_iter=5, verbose=False):
        self.convergence_iter = convergence_iter
        self.verbose = verbose

    def __call__(self, current_info):

        if not hasattr(self, "st"):
            self.st = time.perf_counter()

        current_iter = current_info.gb_info["reader"].iteration

        if current_iter > self.convergence_iter:

            nleaves_cc = curr.gb_info["reader"].get_nleaves()["gb"][:, 0]

            # do not include most recent
            nleaves_cc_max_old = nleaves_cc[:-self.convergence_iter].max()
            nleaves_cc_max_new = nleaves_cc[-self.convergence_iter:].max()

            if nleaves_cc_max_old > nleaves_cc_max_new:
                stop = True

            else:
                stop = False

            if self.verbose:
                dur = (time.perf_counter() - self.st) / 3600.0  # hours
                print(
                    "\nnleaves max old:\n",
                    nleaves_cc_max_old,
                    "\nnleaves max new:\n",
                    nleaves_cc_max_newf,
                    f"\nTIME TO NOW: {dur} hours"
                )

        return stop


class SearchConvergeStopping(Stopping):
    def __init__(self, n_iters=30, diff=1.0, verbose=False, start_iteration=0):
        self.n_iters = n_iters
        self.iters_consecutive = 0
        self.past_like_best = -np.inf
        self.diff = diff
        self.verbose = verbose
        self.start_iteration = start_iteration

    def __call__(self, iter, sample, sampler):

        like_best = sampler.get_log_like(discard=self.start_iteration).max()

        if np.abs(like_best - self.past_like_best) < self.diff:
            self.iters_consecutive += 1

        else:
            self.iters_consecutive = 0
            self.past_like_best = like_best

        if self.verbose:
            print(
                "\nITERS CONSECUTIVE:\n",
                self.iters_consecutive,
                self.past_like_best,
                like_best,
            )

        if self.iters_consecutive >= self.n_iters:
            self.iters_consecutive = 0
            return True

        else:
            return False



class GBBandLogLConvergeStopping(Stopping):

    def __init__(self, fd, band_edges, n_iters=30, diff=1.0, verbose=False, start_iteration=0):
        self.band_edge_inds = np.searchsorted(fd, band_edges, side="right") - 1
        self.num_bands = self.band_edge_inds.shape[0] - 1
        self.converged = np.zeros(self.num_bands, dtype=bool)
        self.iters_consecutive = np.zeros(self.num_bands, dtype=int)
        self.past_like_best = np.full(self.num_bands, -np.inf)
        self.n_iters = n_iters
        self.diff = diff
        self.verbose = verbose
        self.start_iteration = start_iteration

    def add_mgh(self, mgh):
        self.mgh = mgh

    def __call__(self, i, sample, sampler):
        
        ll_per_band = self.mgh.get_ll(band_edge_inds=self.band_edge_inds).max(axis=0)
        
        ll_movement = (ll_per_band - self.past_like_best) > self.diff

        self.iters_consecutive[~ll_movement] += 1
        self.iters_consecutive[ll_movement] = 0

        self.converged = self.iters_consecutive >= self.n_iters

        self.past_like_best[ll_movement] = ll_per_band[ll_movement]

        # for move in sampler.all_moves:
        #     move.converged_sub_bands = self.converged.copy()

        if self.verbose:
            print("Num still going:", (~self.converged).sum(), "\nChanged here:", (ll_movement).sum())

        if np.all(self.converged):
            return True
        else:
            return False

        

            


class SearchConvergeStopping2(Stopping):
    def __init__(self, n_iters=30, diff=0.1, verbose=False, start_iteration=0, iter_back_check=-1):
        self.n_iters = n_iters
        self.iters_consecutive = 0
        self.past_like_best = -np.inf
        self.diff = diff
        self.verbose = verbose
        self.start_iteration = start_iteration
        self.iter_back_check = iter_back_check
        self.time = 0
        self.back_check = [None for _ in range(self.iter_back_check)]
        self.last_sampler_iteration = self.start_iteration
        self.back_check_ind = 0
        self.stop_here = True

    def __call__(self, iter, sample, sampler):

        self.time += 1

        if sampler.iteration <= self.start_iteration:
            return False

        lps = sampler.get_log_like(discard=self.start_iteration)[self.last_sampler_iteration - self.start_iteration:]
        try:
           like_best = lps.max()
        except:
            breakpoint()
        self.last_sampler_iteration = sampler.iteration
        
        if np.any(np.asarray(self.back_check) == None):
            for i in range(len(self.back_check)):
                if self.back_check[i] is None:
                    self.back_check[i] = like_best
                    return False

        first_check = like_best - self.past_like_best > self.diff
        second_check = np.all(like_best >= np.asarray(self.back_check))

        # spread in stored values is below difference
        third_check = np.asarray(self.back_check).max() - np.asarray(self.back_check).min() < self.diff

        update = (
            (first_check and second_check and self.past_like_best == -np.inf) 
            or (self.past_like_best == -np.inf and third_check)
            or (self.past_like_best > -np.inf and first_check)
        )

        self.back_check[self.back_check_ind] = like_best
        self.back_check_ind = (self.back_check_ind + 1) % len(self.back_check)
        
        if update:
            self.past_like_best = like_best
            self.iters_consecutive = 0

        elif self.past_like_best > -np.inf:
            self.iters_consecutive += 1

        if self.verbose:
            print(
                "\nITERS CONSECUTIVE:\n",
                self.iters_consecutive,
                f"previous best: {self.past_like_best}, overall best: {like_best},",
                "first check:", first_check,
                "second check:", second_check
            )

        if self.iters_consecutive >= self.n_iters:
            self.iters_consecutive = 0
            return True

        else:
            return False



class EvidenceStopping(Stopping):
    def __init__(self, diff=0.5, verbose=False):
        self.diff = diff
        self.verbose = verbose

    def __call__(self, iter, sample, sampler):

        betas = sampler.get_betas()[-1]
        logls = sampler.get_log_like().mean(axis=(0, 2))

        logZ, dlogZ = thermodynamic_integration_log_evidence(betas, logls)
        print(logZ, dlogZ)
        return False
        

        if self.verbose:
            print(
                "snr_best",
                snr_best,
                "limit:",
                self.snr_limit,
                "loglike:",
                log_best,
                # d_h_best,
                # h_h_best,
            )

        if snr_best > self.snr_limit:
            return True

        else:
            return False


class MPICommunicateStopping(Stopping):

    def __init__(self, stopper_rank, other_ranks, stop_fn=None):

        self.stopper_rank = stopper_rank
        self.other_ranks = other_ranks
        self.stop_fn = stop_fn

    def add_comm(self, comm):
        self.comm = comm

    def __call__(self, *args, **kwargs):

        if not hasattr(self, "comm"):
            raise ValueError("Must add comm via add_comm method before __call__ is used.")

        if not hasattr(self, "rank"):
            self.rank = self.comm.Get_rank()
            if not self.rank == self.stopper_rank and not self.rank in self.other_ranks:
                raise ValueError("Rank is not available in other ranks list. Must be either stopper rank or in other ranks list.")

            if self.stopper_rank == self.rank and self.stop_fn is None:
                raise ValueError("Rank is equivalent to stopper rank but stop_fn is not provided. It must be provided.")

        if self.rank == self.stopper_rank:
            stop = self.stop_fn(*args, **kwargs)

            if stop:
                for rank in self.other_ranks:
                    tag = int(str(rank) + "1000")
                    self.comm.isend(True, dest=rank, tag=tag)

        else:
            tag = int(str(self.rank) + "1000")
            check_stop = self.comm.irecv(source=self.stopper_rank, tag=tag)

            if check_stop.get_status():
                stop = check_stop.wait()

            else:
                check_stop.cancel()
                stop = False
        
        return stop




        
